#! /usr/bin/env python
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys

from ginga import colors
import ginga.toolkit as ginga_toolkit
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.util.toolbox import ModeIndicator
from ginga.misc import log
from ginga.util.loader import load_data


class FitsViewer(object):

    def __init__(self, logger):
        self.logger = logger
        self.drawcolors = colors.get_colors()
        self.dc = get_canvas_types()

        from ginga.gw import Widgets, Viewers

        self.app = Widgets.Application(logger=logger)
        #self.app.add_callback('shutdown', self.quit)
        self.top = self.app.make_window("Ginga shared canvas example")
        self.top.add_callback('close', self.closed)

        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(1)

        hbox = Widgets.HBox()
        hbox.set_border_width(2)
        hbox.set_spacing(4)

        v1 = Viewers.CanvasView(logger)
        v1.enable_autocuts('on')
        v1.set_autocut_params('zscale')
        v1.enable_autozoom('on')
        v1.set_zoom_algorithm('rate')
        v1.set_zoomrate(1.4)
        v1.show_pan_mark(True)
        v1.set_callback('drag-drop', self.drop_file)
        v1.set_callback('none-move', self.motion)
        v1.set_bg(0.2, 0.2, 0.2)
        v1.ui_set_active(True)
        v1.set_name('tweedledee')
        self.viewer1 = v1
        self._mi1 = ModeIndicator(v1)

        bd = v1.get_bindings()
        bd.enable_all(True)

        # shared canvas between the two viewers
        shcanvas = self.dc.DrawingCanvas()
        # Tell viewer1 to use this canvas
        v1.set_canvas(shcanvas)

        v1.set_desired_size(300, 300)
        iw = Viewers.GingaViewerWidget(viewer=v1)
        hbox.add_widget(iw, stretch=1)

        # Add a second viewer viewing the same canvas
        v2 = Viewers.CanvasView(logger)
        v2.enable_autocuts('on')
        v2.set_autocut_params('zscale')
        v2.enable_autozoom('on')
        v2.set_zoom_algorithm('rate')
        v2.set_zoomrate(1.4)
        v2.show_pan_mark(True)
        v2.set_callback('drag-drop', self.drop_file)
        v2.set_callback('none-move', self.motion)
        v2.set_bg(0.2, 0.2, 0.2)
        v2.ui_set_active(True)
        v1.set_name('tweedledum')
        self.viewer2 = v2
        self._mi2 = ModeIndicator(v2)

        # Tell viewer2 to use this same canvas
        v2.set_canvas(shcanvas)

        bd = v2.get_bindings()
        bd.enable_all(True)

        v2.set_desired_size(300, 300)
        iw = Viewers.GingaViewerWidget(viewer=v2)
        hbox.add_widget(iw, stretch=1)

        # 2nd canvas as a subcanvas of the shared canvas
        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.register_for_cursor_drawing(v1)
        canvas.register_for_cursor_drawing(v2)
        canvas.set_drawtype('rectangle', color='lightblue')
        self.canvas = canvas
        shcanvas.add(self.canvas)
        shcanvas.ui_set_active(True)
        canvas.ui_set_active(True)
        canvas.set_surface(v1)

        self.drawtypes = canvas.get_drawtypes()
        self.drawtypes.sort()

        vbox.add_widget(hbox, stretch=1)

        self.readout = Widgets.Label("")
        vbox.add_widget(self.readout, stretch=0)

        hbox = Widgets.HBox()
        hbox.set_border_width(2)

        wdrawtype = Widgets.ComboBox()
        for name in self.drawtypes:
            wdrawtype.append_text(name)
        index = self.drawtypes.index('rectangle')
        wdrawtype.set_index(index)
        wdrawtype.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wdrawtype = wdrawtype

        wdrawcolor = Widgets.ComboBox()
        for name in self.drawcolors:
            wdrawcolor.append_text(name)
        index = self.drawcolors.index('lightblue')
        wdrawcolor.set_index(index)
        wdrawcolor.add_callback('activated', lambda w, idx: self.set_drawparams())
        self.wdrawcolor = wdrawcolor

        wfill = Widgets.CheckBox("Fill")
        wfill.add_callback('activated', lambda w, tf: self.set_drawparams())
        self.wfill = wfill

        walpha = Widgets.SpinBox(dtype=float)
        walpha.set_limits(0.0, 1.0, incr_value=0.1)
        walpha.set_value(1.0)
        walpha.set_decimals(2)
        walpha.add_callback('value-changed', lambda w, val: self.set_drawparams())
        self.walpha = walpha

        wclear = Widgets.Button("Clear Canvas")
        wclear.add_callback('activated', lambda w: self.clear_canvas())
        wopen = Widgets.Button("Open File")
        wopen.add_callback('activated', lambda w: self.open_file())
        wquit = Widgets.Button("Quit")
        wquit.add_callback('activated', lambda w: self.quit())

        hbox.add_widget(Widgets.Label(''), stretch=1)
        for w in (wopen, wdrawtype, wdrawcolor, wfill,
                  Widgets.Label('Alpha:'), walpha, wclear, wquit):
            hbox.add_widget(w, stretch=0)

        vbox.add_widget(hbox, stretch=0)

        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        btn1 = Widgets.RadioButton("Draw")
        btn1.set_state(mode == 'draw')
        btn1.add_callback('activated', lambda w, val: self.set_mode_cb('draw', val))
        btn1.set_tooltip("Choose this to draw on the canvas")
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Edit", group=btn1)
        btn2.set_state(mode == 'edit')
        btn2.add_callback('activated', lambda w, val: self.set_mode_cb('edit', val))
        btn2.set_tooltip("Choose this to edit things on the canvas")
        hbox.add_widget(btn2)

        ## btn3 = Widgets.CheckBox("I'm using a trackpad")
        ## btn3.add_callback('activated', lambda w, tf: self.use_trackpad_cb(tf))
        ## hbox.add_widget(btn3)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        vbox.add_widget(hbox, stretch=0)

        self.top.set_widget(vbox)

    def set_drawparams(self):
        index = self.wdrawtype.get_index()
        kind = self.drawtypes[index]
        index = self.wdrawcolor.get_index()
        fill = self.wfill.get_state()
        alpha = self.walpha.get_value()
        coord = 'data'

        params = {'color': self.drawcolors[index],
                  'alpha': alpha,
                  'coord': coord,
                  }
        if kind in ('circle', 'rectangle', 'polygon', 'triangle',
                    'righttriangle', 'ellipse', 'square', 'box'):
            params['fill'] = fill
            params['fillalpha'] = alpha

        self.canvas.set_drawtype(kind, **params)

    def clear_canvas(self):
        self.canvas.delete_all_objects()

    def load_file(self, viewer, filepath):
        image = load_data(filepath, logger=self.logger)
        viewer.set_image(image)
        self.top.set_title(filepath)

    def open_file(self):
        from ginga.gw import Widgets

        res = Widgets.FileDialog.getOpenFileName(self, "Open FITS file",
                                                 ".", "FITS files (*.fits)")
        if isinstance(res, tuple):
            fileName = res[0]
        else:
            fileName = str(res)
        if len(fileName) != 0:
            self.load_file(self.viewer1, fileName)

    def drop_file(self, viewer, paths):
        fileName = paths[0]
        #print(fileName)
        self.load_file(viewer, fileName)

    def motion(self, viewer, button, data_x, data_y):

        # Get the value under the data coordinates
        try:
            #value = viewer.get_data(data_x, data_y)
            # We report the value across the pixel, even though the coords
            # change halfway across the pixel
            value = viewer.get_data(int(data_x + 0.5), int(data_y + 0.5))

        except Exception:
            value = None

        fits_x, fits_y = data_x + 1, data_y + 1

        # Calculate WCS RA
        try:
            # NOTE: image function operates on DATA space coords
            image = viewer.get_image()
            if image is None:
                # No image loaded
                return
            ra_txt, dec_txt = image.pixtoradec(fits_x, fits_y,
                                               format='str', coords='fits')
        except Exception as e:
            self.logger.warning("Bad coordinate conversion: %s" % (
                str(e)))
            ra_txt = 'BAD WCS'
            dec_txt = 'BAD WCS'

        text = "RA: %s  DEC: %s  X: %.2f  Y: %.2f  Value: %s" % (
            ra_txt, dec_txt, fits_x, fits_y, value)
        self.readout.set_text(text)

    def set_mode_cb(self, mode, tf):
        self.logger.info("canvas mode changed (%s) %s" % (mode, tf))
        if not (tf is False):
            self.canvas.set_draw_mode(mode)
        return True

    def closed(self, w):
        self.logger.info("Top window closed.")
        self.top = None
        sys.exit()

    def quit(self, *args):
        self.logger.info("Attempting to shut down the application...")
        if self.top is not None:
            self.top.close()
        sys.exit()

    def mainloop(self):
        while True:
            self.app.process_events()


def main(options, args):

    logger = log.get_logger("example2", options=options)

    if options.toolkit is None:
        logger.error("Please choose a GUI toolkit with -t option")

    # decide our toolkit, then import
    ginga_toolkit.use(options.toolkit)

    viewer = FitsViewer(logger)

    viewer.top.resize(700, 540)

    if len(args) > 0:
        viewer.load_file(viewer.viewer1, args[0])

    viewer.top.show()
    viewer.top.raise_()

    try:
        app = viewer.top.get_app()
        app.mainloop()

    except KeyboardInterrupt:
        print("Terminating viewer...")
        if viewer.top is not None:
            viewer.top.close()


if __name__ == "__main__":

    # Parse command line options
    from argparse import ArgumentParser

    argprs = ArgumentParser()

    argprs.add_argument("--debug", dest="debug", default=False,
                        action="store_true",
                        help="Enter the pdb debugger on main()")
    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt)")
    argprs.add_argument("--profile", dest="profile", action="store_true",
                        default=False,
                        help="Run the profiler on main()")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    # Are we debugging this?
    if options.debug:
        import pdb

        pdb.run('main(options, args)')

    # Are we profiling this?
    elif options.profile:
        import profile

        print(("%s profile:" % sys.argv[0]))
        profile.run('main(options, args)')

    else:
        main(options, args)

# END



#
# addons.py -- Goodies for enhancing a Ginga viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#


def show_pan_mark(viewer, tf, color='red'):
    """Show a mark in the pan position (center of window).

    Parameters
    ----------
    viewer : an ImageView subclass instance
        If True, show the color bar; else remove it if present.

    tf : bool
        If True, show the mark; else remove it if present.

    color : str
        Color of the mark; default is 'red'.
    """
    tag = '_$pan_mark'
    radius = 10

    canvas = viewer.get_private_canvas()
    try:
        mark = canvas.get_object_by_tag(tag)
        if not tf:
            canvas.delete_object_by_tag(tag)
        else:
            mark.color = color

    except KeyError:
        if tf:
            Point = canvas.get_draw_class('point')
            canvas.add(Point(0, 0, radius, style='plus', color=color,
                             coord='cartesian'),
                       tag=tag, redraw=False)

    canvas.update_canvas(whence=3)


def show_mode_indicator(viewer, tf, corner='ur'):
    """Show a keyboard mode indicator in one of the corners.

    Parameters
    ----------
    viewer : an ImageView subclass instance
        If True, show the color bar; else remove it if present.

    tf : bool
        If True, show the mark; else remove it if present.

    corner : str
        One of 'll', 'lr', 'ul' or 'ur' selecting a corner.
        The default is 'ur'.

    """
    tag = '_$mode_indicator'

    canvas = viewer.get_private_canvas()
    try:
        indic = canvas.get_object_by_tag(tag)
        if not tf:
            canvas.delete_object_by_tag(tag)
        else:
            indic.corner = corner

    except KeyError:
        if tf:
            # force a redraw if the mode changes
            bm = viewer.get_bindmap()
            bm.add_callback('mode-set',
                            lambda *args: viewer.redraw(whence=3))

            Indicator = canvas.get_draw_class('modeindicator')
            canvas.add(Indicator(corner=corner),
                       tag=tag, redraw=False)

    canvas.update_canvas(whence=3)


def show_color_bar(viewer, tf, side='bottom'):
    """Show a color bar in the window.

    Parameters
    ----------
    viewer : an ImageView subclass instance
        If True, show the color bar; else remove it if present.

    tf : bool
        If True, show the color bar; else remove it if present.

    side : str
        One of 'top' or 'bottom'. The default is 'bottom'.

    """

    tag = '_$color_bar'
    canvas = viewer.get_private_canvas()
    try:
        cbar = canvas.get_object_by_tag(tag)
        if not tf:
            canvas.delete_object_by_tag(tag)
        else:
            cbar.side = side

    except KeyError:
        if tf:
            Cbar = canvas.get_draw_class('colorbar')
            canvas.add(Cbar(side=side), tag=tag, redraw=False)

    canvas.update_canvas(whence=3)


def show_focus_indicator(viewer, tf, color='white'):
    """Show a focus indicator in the window.

    Parameters
    ----------
    viewer : an ImageView subclass instance
        If True, show the color bar; else remove it if present.

    tf : bool
        If True, show the color bar; else remove it if present.

    color : str
        Color for the focus indicator.

    """

    tag = '_$focus_indicator'
    canvas = viewer.get_private_canvas()
    try:
        fcsi = canvas.get_object_by_tag(tag)
        if not tf:
            canvas.delete_object_by_tag(tag)
        else:
            fcsi.color = color

    except KeyError:
        if tf:
            Fcsi = canvas.get_draw_class('focusindicator')
            fcsi = Fcsi(color=color)
            canvas.add(fcsi, tag=tag, redraw=False)
            viewer.add_callback('focus', fcsi.focus_cb)

    canvas.update_canvas(whence=3)


def add_zoom_buttons(viewer, canvas=None, color='black'):
    """Add zoom buttons to a canvas.

    Parameters
    ----------
    viewer : an ImageView subclass instance
        If True, show the color bar; else remove it if present.

    canvas : a DrawingCanvas instance
        The canvas to which the buttons should be added.  If not supplied
        defaults to the private canvas of the viewer.

    color : str
        A color name, hex triplet. The default is 'black'.

    """
    def zoom(box, canvas, event, pt, viewer, n):
        zl = viewer.get_zoom()
        zl += n
        if zl == 0.0:
            zl += n
        viewer.zoom_to(zl + n)

    def add_buttons(viewer, canvas, tag):
        objs = []
        wd, ht = viewer.get_window_size()
        SquareBox = canvas.get_draw_class('squarebox')
        Text = canvas.get_draw_class('text')
        Compound = canvas.get_draw_class('compoundobject')
        x1, y1 = wd - 20, ht // 2 + 20
        zoomin = SquareBox(x1, y1, 15, color='yellow', fill=True,
                           fillcolor='gray', fillalpha=0.5, coord='window')
        zoomin.editable = False
        zoomin.pickable = True
        zoomin.add_callback('pick-down', zoom, viewer, 1)
        objs.append(zoomin)
        x2, y2 = wd - 20, ht // 2 - 20
        zoomout = SquareBox(x2, y2, 15, color='yellow', fill=True,
                            fillcolor='gray', fillalpha=0.5, coord='window')
        zoomout.editable = False
        zoomout.pickable = True
        zoomout.add_callback('pick-down', zoom, viewer, -1)
        objs.append(zoomout)
        objs.append(Text(x1 - 4, y1 + 6, text='+', fontsize=18, color=color,
                         coord='window'))
        objs.append(Text(x2 - 4, y2 + 6, text='--', fontsize=18, color=color,
                         coord='window'))
        obj = Compound(*objs)
        obj.opaque = False
        canvas.add(obj, tag=tag)

    def zoom_resize(viewer, width, height, canvas, tag):
        try:
            canvas.get_object_by_tag(tag)
        except KeyError:
            return False

        canvas.delete_object_by_tag(tag)
        add_buttons(viewer, canvas, tag)

    tag = '_$zoom_buttons'
    if canvas is None:
        canvas = viewer.get_private_canvas()
    canvas.ui_set_active(True)
    canvas.register_for_cursor_drawing(viewer)
    canvas.set_draw_mode('pick')
    viewer.add_callback('configure', zoom_resize, canvas, tag)

    add_buttons(viewer, canvas, tag)


# END



#
# ImageViewJpw.py -- Module for a Ginga FITS viewer in a Jupyter web notebook.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
This example illustrates using a Ginga as the driver of a Jupyter web widget.

REQUIREMENTS:
  To use this code you will need the "ipywidgets" and "ipyevents" python
  modules installed.  These are easily installed via:

     $ pip install ipyevents
     $ jupyter nbextension enable --py --sys-prefix ipyevents

  or via conda:

     $ conda install -c conda-forge ipyevents

Basic usage in a Jupyter notebook:

import ipywidgets as widgets

# create a Jupyter image that will be our display surface
# format can be 'jpeg' or 'png'; specify width and height to set viewer size
jp_img = widgets.Image(format='jpeg', width=500, height=500)

# Boilerplate to create a Ginga viewer connected to this widget
# this could be simplified by creating a class that created viewers
# as a factory.
from ginga.misc.log import get_logger
logger = get_logger("v1", log_stderr=True, level=40)

from ginga.web.jupyterw.ImageViewJpw import EnhancedCanvasView
v1 = EnhancedCanvasView(logger=logger)
v1.set_widget(jp_img)

bd = v1.get_bindings()
bd.enable_all(True)

# You can now build a GUI with the image widget and other Jupyter
# widgets.  Here we just show the image widget.
v1.embed()

"""
from ipyevents import Event as EventListener

from ginga import ImageView, AstroImage
from ginga import Mixins, Bindings
from ginga.canvas import render
from ginga.util.toolbox import ModeIndicator
from ginga.util import loader

from ginga.web.jupyterw import JpHelp


class ImageViewJpwError(ImageView.ImageViewError):
    pass


class ImageViewJpw(ImageView.ImageViewBase):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageView.ImageViewBase.__init__(self, logger=logger,
                                         rgbmap=rgbmap,
                                         settings=settings)

        self.t_.set_defaults(renderer='cairo')

        self.rgb_order = 'RGBA'
        self.jp_img = None
        self.jp_evt = None

        self._defer_task = None
        self.msgtask = None

        self.renderer = None
        # Pick a renderer that can work with us
        renderers = ['cairo', 'agg', 'pil', 'opencv']
        preferred = self.t_['renderer']
        if preferred in renderers:
            renderers.remove(preferred)
        self.possible_renderers = [preferred] + renderers
        self.choose_best_renderer()

    def set_widget(self, jp_img):
        """Call this method with the Jupyter image widget (image_w)
        that will be used.
        """
        self.jp_img = jp_img

        # TODO: need configure (resize) event callback

        # see reschedule_redraw() method
        self._defer_task = JpHelp.Timer()
        self._defer_task.add_callback('expired',
                                      lambda timer: self.delayed_redraw())
        self.msgtask = JpHelp.Timer()
        self.msgtask.add_callback('expired',
                                  lambda timer: self.onscreen_message(None))

        # for some reason these are stored as strings!
        wd, ht = int(jp_img.width), int(jp_img.height)
        self.configure_window(wd, ht)

    def get_widget(self):
        return self.jp_img

    def choose_renderer(self, name):
        klass = render.get_render_class(name)
        self.renderer = klass(self)

        if self.jp_img is not None:
            wd, ht = int(self.jp_img.width), int(self.jp_img.height)
            self.configure_window(wd, ht)

    def choose_best_renderer(self):
        for name in self.possible_renderers:
            try:
                self.choose_renderer(name)
                self.logger.info("best renderer available is '{}'".format(name))
                return
            except Exception as e:
                continue

        raise ImageViewJpwError("No valid renderers available: {}".format(str(self.possible_renderers)))

    def update_image(self):
        fmt = self.jp_img.format
        web_img = self.renderer.get_surface_as_rgb_format_bytes(
            format=fmt)
        # this updates the model, and then the Jupyter image(s)
        self.jp_img.value = web_img

    def reschedule_redraw(self, time_sec):
        self._defer_task.stop()
        self._defer_task.start(time_sec)

    def configure_window(self, width, height):
        self.configure_surface(width, height)

    def _resize_cb(self, event):
        self.configure_window(event.width, event.height)

    def set_cursor(self, cursor):
        # TODO
        pass

    def onscreen_message(self, text, delay=None, redraw=True):
        if self.jp_img is None:
            return
        self.msgtask.stop()
        self.set_onscreen_message(text, redraw=redraw)
        if delay is not None:
            self.msgtask.start(delay)


class ImageViewEvent(ImageViewJpw):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageViewJpw.__init__(self, logger=logger, rgbmap=rgbmap,
                              settings=settings)

        self._button = 0

        # maps EventListener events to callback handlers
        self._evt_dispatch = {
            'mousedown': self.button_press_event,
            'mouseup': self.button_release_event,
            'mousemove': self.motion_notify_event,
            'wheel': self.scroll_event,
            'mouseenter': self.enter_notify_event,
            'mouseleave': self.leave_notify_event,
            'keydown': self.key_press_event,
            'keyup': self.key_release_event,
        }

        # mapping from EventListener key events to ginga key events
        self._keytbl = {
            'shiftleft': 'shift_l',
            'shiftright': 'shift_r',
            'controlleft': 'control_l',
            'controlright': 'control_r',
            'altleft': 'alt_l',
            'altright': 'alt_r',
            'osleft': 'super_l',
            'osright': 'super_r',
            'contextmenu': 'menu_r',
            'backslash': 'backslash',
            'space': 'space',
            'escape': 'escape',
            'enter': 'return',
            'tab': 'tab',
            'arrowright': 'right',
            'arrowleft': 'left',
            'arrowup': 'up',
            'arrowdown': 'down',
            'pageup': 'page_up',
            'pagedown': 'page_down',
            'f1': 'f1',
            'f2': 'f2',
            'f3': 'f3',
            'f4': 'f4',
            'f5': 'f5',
            'f6': 'f6',
            'f7': 'f7',
            'f8': 'f8',
            'f9': 'f9',
            'f10': 'f10',
            'f11': 'f11',
            'f12': 'f12',
        }

        self._keytbl2 = {
            '`': 'backquote',
            '"': 'doublequote',
            "'": 'singlequote',
        }

        # Define cursors for pick and pan
        #hand = openHandCursor()
        hand = 'fleur'
        self.define_cursor('pan', hand)
        cross = 'cross'
        self.define_cursor('pick', cross)

        for name in ('motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     'pinch', 'rotate', 'pan', 'swipe', 'tap'):
            self.enable_callback(name)

    def set_widget(self, jp_imgw):
        """Call this method with the Jupyter image widget (image_w)
        that will be used.
        """
        super(ImageViewEvent, self).set_widget(jp_imgw)

        self.jp_evt = EventListener(source=jp_imgw)
        self.jp_evt.watched_events = [
            'keydown', 'keyup', 'mouseenter', 'mouseleave',
            'mousedown', 'mouseup', 'mousemove', 'wheel',
            'contextmenu'
        ]
        self.jp_evt.prevent_default_action = True

        self.jp_evt.on_dom_event(self._handle_event)
        self.logger.info("installed event handlers")

        return self.make_callback('map')

    def _handle_event(self, event):
        # TODO: need focus events and maybe a map event
        # TODO: Set up widget as a drag and drop destination
        evt_kind = event['type']
        handler = self._evt_dispatch.get(evt_kind, None)
        if handler is not None:
            return handler(event)
        return False

    def transkey(self, keycode, keyname=None):
        keycode = str(keycode).lower()
        if keyname is None:
            keyname = keycode
        self.logger.debug("key code in jupyter '%s'" % (keycode))
        res = self._keytbl.get(keycode, None)
        if res is None:
            res = self._keytbl2.get(keyname, keyname)
        return res

    def get_key_table(self):
        return self._keytbl

    def focus_event(self, event, has_focus):
        return self.make_callback('focus', has_focus)

    def enter_notify_event(self, event):
        enter_focus = self.t_.get('enter_focus', False)
        if enter_focus:
            # TODO: set focus on canvas
            pass
        return self.make_callback('enter')

    def leave_notify_event(self, event):
        self.logger.debug("leaving widget...")
        return self.make_callback('leave')

    def key_press_event(self, event):
        keyname = self.transkey(event['code'], keyname=event['key'])
        self.logger.debug("key press event, key=%s" % (keyname))
        return self.make_ui_callback('key-press', keyname)

    def key_release_event(self, event):
        keyname = self.transkey(event['code'], keyname=event['key'])
        self.logger.debug("key release event, key=%s" % (keyname))
        return self.make_ui_callback('key-release', keyname)

    def button_press_event(self, event):
        x, y = event['dataX'], event['dataY']
        self.last_win_x, self.last_win_y = x, y

        button = 0
        button |= 0x1 << event['button']
        self._button = button
        self.logger.debug("button event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('button-press', button, data_x, data_y)

    def button_release_event(self, event):
        x, y = event['dataX'], event['dataY']
        self.last_win_x, self.last_win_y = x, y

        button = 0
        button |= 0x1 << event['button']
        self._button = 0
        self.logger.debug("button release at %dx%d button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('button-release', button, data_x, data_y)

    def motion_notify_event(self, event):
        button = self._button
        x, y = event['dataX'], event['dataY']
        self.last_win_x, self.last_win_y = x, y

        self.logger.debug("motion event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('motion', button, data_x, data_y)

    def scroll_event(self, event):
        x, y = event['dataX'], event['dataY']
        self.last_win_x, self.last_win_y = x, y
        dx, dy = event['deltaX'], event['deltaY']

        if (dx != 0 or dy != 0):
            # <= This browser gives us deltas for x and y
            # Synthesize this as a pan gesture event
            self.make_ui_callback('pan', 'start', 0, 0)
            self.make_ui_callback('pan', 'move', -dx, -dy)
            return self.make_ui_callback('pan', 'stop', 0, 0)

        # <= This code path should not be followed under normal
        # circumstances.
        # we leave it here in case we want to make the scroll
        # callback configurable in the future

        # TODO: calculate actual angle of direction
        if dy < 0:
            direction = 0.0   # up
        elif dy > 0:
            direction = 180.0   # down
        else:
            return False

        # 15 deg is standard 1-click turn for a wheel mouse
        num_deg = 15.0
        self.logger.debug("scroll deg=%f direction=%f" % (
            num_deg, direction))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('scroll', direction, num_deg,
                                     data_x, data_y)


class ImageViewZoom(Mixins.UIMixin, ImageViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 bindmap=None, bindings=None):
        ImageViewEvent.__init__(self, logger=logger, rgbmap=rgbmap,
                                settings=settings)
        Mixins.UIMixin.__init__(self)

        self.ui_set_active(True)

        if bindmap is None:
            bindmap = ImageViewZoom.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = ImageViewZoom.bindingsClass(self.logger)
        self.set_bindings(bindings)

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)


class CanvasView(ImageViewZoom):

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger, settings=settings,
                               rgbmap=rgbmap,
                               bindmap=bindmap, bindings=bindings)

        # Needed for UIMixin to propagate events correctly
        self.objects = [self.private_canvas]

        self._mi = ModeIndicator(self)

    def set_canvas(self, canvas, private_canvas=None):
        super(CanvasView, self).set_canvas(canvas,
                                           private_canvas=private_canvas)

        self.objects[0] = self.private_canvas


class EnhancedCanvasView(CanvasView):
    """
    This just adds some convenience methods to the viewer for loading images,
    grabbing screenshots, etc.  You can subclass to add new methods.
    """

    def embed(self):
        """
        Embed a viewer into a Jupyter notebook.
        """
        return self.jp_img

    def open(self, new=1):
        """
        Open this viewer in a new browser window or tab.
        """
        # TBD
        raise Exception("Not yet implemented!")

    def show(self, fmt=None):
        """
        Capture the window of a viewer.
        """
        # force any delayed redraws
        # TODO: this really needs to be addressed in get_rgb_image_as_bytes()
        # of the various superclasses, as it affects other backends as well
        self.redraw_now()

        from IPython.display import Image

        if fmt is None:
            # what format are we using for the Jupyter image--use that
            fmt = self.jp_img.format

        return Image(data=bytes(self.get_rgb_image_as_bytes(format=fmt)),
                     format=fmt, embed=True)

    def load(self, filepath):
        """
        Load a file into the viewer.
        """
        image = loader.load_file(filepath, logger=self.logger)

        self.set_image(image)

    load_fits = load

    def load_hdu(self, hdu):
        """
        Load an HDU into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.load_hdu(hdu)

        self.set_image(image)

    def load_data(self, data_np):
        """
        Load raw numpy data into the viewer.
        """
        image = AstroImage.AstroImage(logger=self.logger)
        image.set_data(data_np)

        self.set_image(image)

    def add_canvas(self, tag=None):
        # add a canvas to the view
        my_canvas = self.get_canvas()
        DrawingCanvas = my_canvas.get_draw_class('drawingcanvas')
        canvas = DrawingCanvas()
        # enable drawing on the canvas
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype(None)
        canvas.ui_set_active(True)
        canvas.set_surface(self)
        canvas.register_for_cursor_drawing(self)
        # add the canvas to the view.
        my_canvas.add(canvas, tag=tag)

        return canvas



#
# toolbox.py -- Goodies for enhancing a Ginga viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

# STDLIB
import io
import os
import warnings

# THIRD-PARTY
from astropy.utils.data import _find_pkg_data_path
from astropy.utils.exceptions import AstropyUserWarning


class ModeIndicator(object):
    """
    This class adds a mode status indicator to a viewer's lower right-hand
    corner.

    Usage:
    Instantiate this class with a Ginga ImageView{Toolkit} object as the
    sole constructor argument.  Save a reference to the mode indicator
    object somewhere so it doesn't get collected.

    NOTE: Please don't use this class--it's likely to be deprecated!
    Instead, use the show_mode_indicator() method provided by the
    ImageView class.
    """

    def __init__(self, viewer):
        self.viewer = viewer

        # set to false to disable
        self.visible = True

        self.fontsize = 12
        self.xpad = 8
        self.ypad = 4
        self.offset = 10
        self.modetbl = dict(locked='L', softlock='S', held='H', oneshot='O')

        # for displaying modal keyboard state
        self.mode_obj = None
        bm = viewer.get_bindmap()
        bm.add_callback('mode-set', self.mode_change_cb)
        viewer.add_callback('configure', self._configure_cb)

    def mode_change_cb(self, bindmap, mode, modetype):
        # delete the old indicator
        obj = self.mode_obj
        self.mode_obj = None
        canvas = self.viewer.get_private_canvas()
        if obj:
            try:
                canvas.delete_object(obj)
            except Exception:
                pass

        if not self.visible:
            return True

        # if not one of the standard modifiers, display the new one
        if mode not in (None, 'ctrl', 'shift'):
            Text = canvas.get_draw_class('text')
            Polygon = canvas.get_draw_class('polygon')
            Compound = canvas.get_draw_class('compoundobject')

            text = mode
            if modetype in self.modetbl:
                text += ' [%s]' % self.modetbl[modetype]

            color = 'cyan' if mode == 'meta' else 'yellow'

            o1 = Text(0, 0, text,
                      fontsize=self.fontsize, color=color, coord='window')

            txt_wd, txt_ht = self.viewer.renderer.get_dimensions(o1)

            box_wd, box_ht = 2 * self.xpad + txt_wd, 2 * self.ypad + txt_ht
            win_wd, win_ht = self.viewer.get_window_size()
            x_base, y_base = win_wd - self.offset - box_wd, win_ht - self.offset - box_ht

            o1.x, o1.y = x_base + self.xpad, y_base + txt_ht + self.ypad

            # yellow text on a black filled rectangle
            cx1, cy1, cx2, cy2 = x_base, y_base, x_base + box_wd, y_base + box_ht
            poly_pts = ((cx1, cy1), (cx2, cy1), (cx2, cy2), (cx1, cy2))
            o2 = Compound(Polygon(poly_pts,
                                  color='black', coord='window',
                                  fill=True, fillcolor='black'),
                          o1)
            self.mode_obj = o2
            canvas.add(o2)

        return True

    def _configure_cb(self, view, width, height):
        # redraw the mode indicator since the window has been resized
        bm = view.get_bindmap()
        mode, modetype = bm.current_mode()
        self.mode_change_cb(bm, mode, modetype)


def trim_prefix(text, nchr):
    """Trim characters off of the beginnings of text lines.

    Parameters
    ----------
    text : str
        The text to be trimmed, with newlines (\n) separating lines

    nchr: int
        The number of spaces to trim off the beginning of a line if
        it starts with that many spaces

    Returns
    -------
    text : str
        The trimmed text
    """
    res = []
    for line in text.split('\n'):
        if line.startswith(' ' * nchr):
            line = line[nchr:]
        res.append(line)

    return '\n'.join(res)


def generate_cfg_example(config_name, cfgpath='examples/configs', **kwargs):
    """Generate config file documentation for a given config name.

    If found, it will be a Python code block of the contents.
    If not found, it will have a generic message that the config
    is not available.

    Parameters
    ----------
    config_name : str
        Config name that is attached to the configuration file.
        This is the same as input for ``prefs.createCategory()``.
        For example, ``'general'``, ``'channel_Image'``, or
        ``'plugin_Zoom'``.

    cfgpath : str
        Where it is within package data.

    kwargs : dict
        Optional keywords for :func:`~astropy.utils.data.get_pkg_data_contents`.

    Returns
    -------
    docstr : str
        Docstring to be inserted into documentation.

    """
    cfgname = config_name + '.cfg'

    try:
        cfgfile = _find_pkg_data_path(os.path.join(cfgpath, cfgname), **kwargs)
        with open(cfgfile) as f:
            cfgdata = f.readlines()
    except Exception as e:
        warnings.warn(str(e), AstropyUserWarning)
        return ''

    homepath = '~'  # Symbol for HOME for doc only, not actual code
    userfile = os.path.join(homepath, '.ginga', cfgname)

    docstring = io.StringIO()
    docstring.write("""It is customizable using ``{0}``, where ``{1}``
is your HOME directory:

.. code-block:: Python

""".format(userfile, homepath))

    for line in cfgdata:
        line = line.strip()

        if len(line) == 0:
            docstring.write('\n')  # Prevent trailing spaces
        else:
            docstring.write('  {0}\n'.format(line))

    return docstring.getvalue()

#END



#
# rgb_cms.py -- RGB color management handling.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import os
import glob
import hashlib

from ginga.misc import Bunch

from . import paths

# How about color management (ICC profile) support?
try:
    import PIL.ImageCms as ImageCms
    have_cms = True
except ImportError:
    have_cms = False

have_pilutil = False
try:
    from scipy.misc import toimage, fromimage
    have_pilutil = True
except ImportError:
    pass

basedir = paths.ginga_home

# Holds profiles
profile = {}
rendering_intent = 'perceptual'
intents = dict(perceptual=0)

# The working profile, if there is one
working_profile = None

# Holds transforms
icc_transform = {}


class ColorManager(object):

    def __init__(self, logger):
        self.logger = logger

    def can_profile(self):
        return have_cms

    def profile_to_working_pil(self, image_pil, kwds, intent=None):

        image = image_pil

        # If we have a working color profile then handle any embedded
        # profile or color space information, if possible
        if not have_cms:
            self.logger.warning(
                "No CMS is installed; leaving image unprofiled.")
            return image

        if not have_profile(working_profile):
            self.logger.warning(
                "No working profile defined; leaving image unprofiled.")
            return image

        out_profile = profile[working_profile].name

        if not os.path.exists(profile[out_profile].path):
            self.logger.warning(
                "Working profile '%s' (%s) not found; leaving image "
                "unprofiled." % (out_profile, profile[out_profile].path))
            return image

        if intent is None:
            intent = rendering_intent

        # Assume sRGB image, unless we learn to the contrary
        in_profile = 'sRGB'
        try:
            if 'icc_profile' in image.info:
                self.logger.debug("image has embedded color profile")
                buf_profile = image.info['icc_profile']

                # Write out embedded profile (if needed)
                prof_md5 = hashlib.md5(buf_profile).hexdigest()
                in_profile = os.path.join("/tmp", "_image_%d_%s.icc" % (
                    os.getpid(), prof_md5))
                if not os.path.exists(in_profile):
                    with open(in_profile, 'wb') as icc_f:
                        icc_f.write(buf_profile)

            # see if there is any EXIF tag about the colorspace
            elif 'ColorSpace' in kwds:
                csp = kwds['ColorSpace']
                #iop = kwds.get('InteroperabilityIndex', None)
                if (csp == 0x2) or (csp == 0xffff):
                    # NOTE: 0xffff is really "undefined" and should be
                    # combined with a test of EXIF tag 0x0001
                    # ('InteropIndex') == 'R03', but PIL _getexif()
                    # does not return the InteropIndex
                    in_profile = 'AdobeRGB'
                    self.logger.debug("hmm..this looks like an AdobeRGB image")

                elif csp == 0x1:
                    self.logger.debug("hmm..this looks like a sRGB image")
                    in_profile = 'sRGB'

                else:
                    self.logger.debug("no color space metadata, assuming this is an sRGB image")

            # if we have a valid profile, try the conversion
            tr_key = get_transform_key(in_profile, out_profile, intent,
                                       None, None, 0)
            if tr_key in icc_transform:
                # We have an in-core transform already for this (faster)
                image = convert_profile_pil_transform(image,
                                                      icc_transform[tr_key],
                                                      inPlace=True)
            else:
                # Convert using profiles on disk (slower)
                if in_profile in profile:
                    in_profile = profile[in_profile].path
                image = convert_profile_pil(image, in_profile,
                                            profile[out_profile].path,
                                            intent)

            self.logger.info("converted from profile (%s) to profile (%s)" % (
                in_profile, profile[out_profile].name))

        except Exception as e:
            self.logger.error("Error converting from embedded color profile: %s" % (str(e)))
            self.logger.warning("Leaving image unprofiled.")

        return image

    def profile_to_working_numpy(self, image_np, kwds, intent=None):

        image_in = toimage(image_np)
        image_out = self.profile_to_working_pil(image_in, kwds,
                                                intent=intent)
        return fromimage(image_out)


# --- Color Management conversion functions ---

def convert_profile_pil(image_pil, inprof_path, outprof_path, intent_name,
                        inPlace=False):
    if not have_cms:
        return image_pil

    image_out = ImageCms.profileToProfile(image_pil, inprof_path,
                                          outprof_path,
                                          renderingIntent=intents[intent_name],
                                          outputMode='RGB', inPlace=inPlace,
                                          flags=0)
    if inPlace:
        return image_pil
    return image_out


def convert_profile_pil_transform(image_pil, transform, inPlace=False):
    if not have_cms:
        return image_pil

    image_out = ImageCms.applyTransform(image_pil, transform, inPlace)
    if inPlace:
        return image_pil
    return image_out


def convert_profile_numpy(image_np, inprof_path, outprof_path, intent_name):
    if (not have_pilutil) or (not have_cms):
        return image_np

    in_image_pil = toimage(image_np)
    out_image_pil = convert_profile_pil(in_image_pil,
                                        inprof_path, outprof_path, intent_name)
    image_out = fromimage(out_image_pil)
    return image_out


def convert_profile_numpy_transform(image_np, transform):
    if (not have_pilutil) or (not have_cms):
        return image_np

    in_image_pil = toimage(image_np)
    convert_profile_pil_transform(in_image_pil, transform, inPlace=True)
    image_out = fromimage(in_image_pil)
    return image_out


def get_transform_key(from_name, to_name, to_intent, proof_name,
                      proof_intent, flags):
    return (from_name, to_name, to_intent, proof_name, proof_intent, flags)


def get_transform(from_name, to_name, to_intent='perceptual',
                  proof_name=None, proof_intent=None,
                  use_black_pt=False):
    global icc_transform

    flags = 0
    if proof_name is not None:
        if hasattr(ImageCms, 'FLAGS'):
            # supporting multiple versions of lcms...sigh..
            flags |= ImageCms.FLAGS['SOFTPROOFING']
        else:
            flags |= ImageCms.SOFTPROOFING
    if use_black_pt:
        if hasattr(ImageCms, 'FLAGS'):
            flags |= ImageCms.FLAGS['BLACKPOINTCOMPENSATION']
        else:
            flags |= ImageCms.BLACKPOINTCOMPENSATION

    key = get_transform_key(from_name, to_name, to_intent, proof_name,
                            proof_intent, flags)

    try:
        output_transform = icc_transform[key]

    except KeyError:
        # try to build transform on the fly
        try:
            if not (proof_name is None):
                output_transform = ImageCms.buildProofTransform(
                    profile[from_name].path,
                    profile[to_name].path,
                    profile[proof_name].path,
                    'RGB', 'RGB',
                    renderingIntent=intents[to_intent],
                    proofRenderingIntent=intents[proof_intent],
                    flags=flags)
            else:
                output_transform = ImageCms.buildTransform(
                    profile[from_name].path,
                    profile[to_name].path,
                    'RGB', 'RGB',
                    renderingIntent=intents[to_intent],
                    flags=flags)

            icc_transform[key] = output_transform

        except Exception as e:
            raise Exception("Failed to build profile transform: %s" % (str(e)))

    return output_transform


def convert_profile_fromto(image_np, from_name, to_name,
                           to_intent='perceptual',
                           proof_name=None, proof_intent=None,
                           use_black_pt=False, logger=None):

    try:
        output_transform = get_transform(from_name, to_name,
                                         to_intent=to_intent,
                                         proof_name=proof_name,
                                         proof_intent=proof_intent,
                                         use_black_pt=use_black_pt)

        out_np = convert_profile_numpy_transform(image_np, output_transform)
        return out_np

    except Exception as e:
        if logger is not None:
            logger.warn("Error converting profile from '%s' to '%s': %s" % (
                from_name, to_name, str(e)))
            logger.warn("Leaving image unprofiled")
        return image_np


def set_rendering_intent(intent):
    """
    Sets the color management attribute rendering intent.

    Parameters
    ----------
    intent: integer
      0: perceptual, 1: relative colorimetric, 2: saturation,
      3: absolute colorimetric
    """
    global rendering_intent
    rendering_intent = intent


# Look up all the profiles the user has available
glob_pat = os.path.join(basedir, "profiles", "*.icc")
for path in glob.glob(glob_pat):
    dirname, filename = os.path.split(path)
    profname, ext = os.path.splitext(filename)
    profile[profname] = Bunch.Bunch(name=profname,
                                    path=os.path.abspath(path))

if have_cms:
    d = dict(absolute_colorimetric=ImageCms.INTENT_ABSOLUTE_COLORIMETRIC,
             perceptual=ImageCms.INTENT_PERCEPTUAL,
             relative_colorimetric=ImageCms.INTENT_RELATIVE_COLORIMETRIC,
             saturation=ImageCms.INTENT_SATURATION)
    intents.update(d)

    # Build transforms for profile conversions for which we have profiles


def have_profile(name):
    return name in profile.keys()


def get_profiles():
    names = list(profile.keys())
    names.sort()
    return names


def get_intents():
    names = list(intents.keys())
    names.sort()
    return names


def set_profile_alias(alias, profname):
    global profile
    profile[alias] = profile[profname]

# END



#
# BokehHelp.py -- help classes for Bokeh drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from ginga import colors


class Pen(object):
    def __init__(self, color='black', linewidth=1, linestyle='solid'):
        self.color = color
        self.linewidth = linewidth
        self.linestyle = linestyle


class Brush(object):
    def __init__(self, color='black', fill=False):
        self.color = color
        self.fill = fill


class Font(object):
    def __init__(self, fontname='sans', fontsize=12.0, color='black'):
        self.fontname = fontname
        self.fontsize = fontsize
        self.color = color


class BokehContext(object):

    def __init__(self, plot):
        self.plot = plot
        self.kwdargs = dict()

    def set_canvas(self, plot):
        self.plot = plot

    def init(self, **kwdargs):
        self.kwdargs = dict()
        self.kwdargs.update(kwdargs)

    def set(self, **kwdargs):
        self.kwdargs.update(kwdargs)

    def update_fill(self, brush):
        if brush is not None:
            # NOTE: alpha is taken care of in brush.color
            self.kwdargs['fill_alpha'] = 1.0
            self.kwdargs['fill_color'] = brush.color
        else:
            self.kwdargs['fill_alpha'] = 0.0

    def update_line(self, pen):
        self.kwdargs['line_color'] = pen.color
        self.kwdargs['line_width'] = pen.linewidth
        if pen.linestyle == 'dash':
            self.kwdargs['line_dash'] = [3, 2, 3, 2]
            self.kwdargs['line_dash_offset'] = 0

    def update_patch(self, pen, brush):
        self.update_line(pen)
        self.update_fill(brush)

    def update_font(self, pen, font):
        self.kwdargs['text_font'] = font.fontname
        # Bokeh expects font size specified as a string
        self.kwdargs['text_font_size'] = "%dpt" % (int(font.fontsize))
        self.kwdargs['text_color'] = pen.color

    def get_color(self, color, alpha):
        if isinstance(color, str):
            r, g, b = colors.lookup_color(color)
        elif isinstance(color, tuple):
            # color is assumed to be a 4-tuple of RGBA values as floats
            # between 0 and 1
            r, g, b = color
        else:
            r, g, b = 1.0, 1.0, 1.0

        # from the Bokeh docs:
        # "a 4-tuple of (r,g,b,a) where r, g, b are integers between
        #  0 and 255 and a is a floating point value between 0 and 1"
        ri, gi, bi = int(255 * r), int(255 * g), int(255 * b)

        return (ri, gi, bi, alpha)

    def get_pen(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        color = self.get_color(color, alpha=alpha)
        return Pen(color=color, linewidth=linewidth, linestyle=linestyle)

    def get_brush(self, color, alpha=1.0):
        color = self.get_color(color, alpha=alpha)
        return Brush(color=color, fill=True)

    def get_font(self, name, size, color, alpha=1.0):
        color = self.get_color(color, alpha=alpha)
        return Font(fontname=name, fontsize=size, color=color)

    def text_extents(self, text, font):
        import matplotlib.textpath as textpath

        # This is not completely accurate because it depends a lot
        # on the renderer used, but that is complicated under Bokeh
        t = textpath.TextPath((0, 0), text, size=font.fontsize,
                              prop=font.fontname)
        bb = t.get_extents()
        wd, ht = bb.width, bb.height
        return (wd, ht)

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
The ``Cursor`` plugin displays a summary line of text that changes as the
user moves the cursor around an image.  In the standard reference viewer
configuration, it appears as a line containing green text just below the
``Colorbar`` plugin.

**Plugin Type: Global**

``Cursor`` is a global plugin.  Only one instance can be opened.

**Usage**

``Cursor`` simply tracks the cursor as it moves around an image and displays
information about the pixel coordinates, WCS coordinates (if available)
and the value of the pixel under the cursor.

There is no associated configuration GUI.

.. note:: Pixel coordinates are affected by the general setting
          "pixel_coords_offset" which can be set in the "general.cfg"
          configuration file for ginga.  The default is value for this
          setting is 1.0, which means pixel coordinates are reported
          from an origin of 1, as per the FITS standard.

"""
import platform

import numpy

from ginga import GingaPlugin, toolkit
from ginga.gw import Readout
from ginga.ImageView import ImageViewNoDataError
from ginga.fonts import font_asst

__all__ = ['Cursor']


class Cursor(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(Cursor, self).__init__(fv)

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Cursor')
        self.settings.add_defaults(font_name=None, font_size=None)
        self.settings.load(onError='silent')

        fv.add_callback('field-info', self.field_info_cb)
        fv.set_callback('channel-change', self.focus_cb)

        self.readout = None
        self.gui_up = False

    def build_gui(self, container):
        readout = Readout.Readout(-1, 24)

        # NOTE: Special hack for certain platforms, otherwise the font
        # on the readout is too small
        macos_ver = platform.mac_ver()[0]

        font_size = self.settings.get('font_size', None)
        if font_size is None:
            if len(macos_ver) > 0:
                # Mac OS X
                font_size = 16
            elif toolkit.get_family().startswith('gtk'):
                # Gtk
                font_size = 11
            else:
                font_size = 11

        font_name = self.settings.get('font_name', None)
        if font_name is None:
            font_name = font_asst.resolve_alias('fixed', 'Courier')
            if len(macos_ver) > 0:
                # Mac OS X
                font_name = 'monaco'

        readout.set_font(font_name, font_size)

        self.readout = readout
        rw = self.readout.get_widget()

        container.add_widget(rw, stretch=0)
        self.gui_up = True

    def readout_config(self, fitsimage, image, readout):
        if (readout is None) or (image is None):
            return True
        self.logger.debug("configuring readout (%s)" % (str(readout)))
        # Configure readout for this image.
        # Get and store the sizes of the fields necessary to display
        # all X, Y coords as well as values.

        try:
            width, height = fitsimage.get_data_size()
        except ImageViewNoDataError as exc:  # table
            self.logger.debug(str(exc))
            return

        # Set size of coordinate areas (4 is "." + precision 3)
        readout.maxx = len(str(width)) + 4
        readout.maxy = len(str(height)) + 4
        minval, maxval = image.get_minmax()
        readout.maxv = max(len(str(minval)), len(str(maxval)))
        return True

    def force_update(self, channel):
        viewer = channel.fitsimage
        data_x, data_y = viewer.get_last_data_xy()
        self.fv.showxy(viewer, data_x, data_y)

    def redo(self, channel, image):
        if not self.gui_up or channel is None:
            return
        self.readout_config(channel.fitsimage, image, self.readout)

        # force an update on an image change, because the WCS
        # may be different, even if the data coords are the same
        self.force_update(channel)

    def change_readout(self, channel, fitsimage):
        self.readout.fitsimage = fitsimage

        image = fitsimage.get_image()
        if image is not None:
            self.readout_config(fitsimage, image, self.readout)
            self.logger.debug("configured readout")

    def focus_cb(self, viewer, channel):
        if not self.gui_up or channel is None:
            return

        self.change_readout(channel, channel.fitsimage)

        # force an update on a channel change, because the WCS
        # may be different, even if the data coords are the same
        self.force_update(channel)

    def start(self):
        channel = self.fv.get_channel_info()
        self.focus_cb(self.fv, channel)

    def stop(self):
        self.readout = None
        self.gui_up = False
        return True

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def field_info_cb(self, viewer, channel, info):
        if not self.gui_up:
            return
        readout = self.readout
        fitsimage = channel.fitsimage

        if readout.fitsimage != fitsimage:
            self.change_readout(channel, fitsimage)

        # If this is a multiband image, then average the values
        # for the readout
        value = info.value
        if isinstance(value, numpy.ndarray):
            avg = numpy.average(value)
            value = avg

        # Update the readout
        px_x = "%.3f" % info.x
        px_y = "%.3f" % info.y
        maxx = max(readout.maxx, len(str(px_x)))
        if maxx > readout.maxx:
            readout.maxx = maxx
        maxy = max(readout.maxy, len(str(px_y)))
        if maxy > readout.maxy:
            readout.maxy = maxy
        maxv = max(readout.maxv, len(str(value)))
        if maxv > readout.maxv:
            readout.maxv = maxv

        if 'ra_txt' in info:
            text = "%1.1s: %-14.14s  %1.1s: %-14.14s  X: %-*.*s  Y: %-*.*s  Value: %-*.*s" % (
                info.ra_lbl, info.ra_txt, info.dec_lbl, info.dec_txt,
                maxx, maxx, px_x, maxy, maxy, px_y, maxv, maxv, value)
        else:
            text = "%1.1s: %-14.14s  %1.1s: %-14.14s  X: %-*.*s  Y: %-*.*s  Value: %-*.*s" % (
                '', '', '', '',
                maxx, maxx, px_x, maxy, maxy, px_y, maxv, maxv, value)
        readout.set_text(text)

    def __str__(self):
        return 'cursor'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Cursor', package='ginga')

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
``Histogram`` plots a histogram for a region drawn in the image, or for the
entire image.

**Plugin Type: Local**

``Histogram`` is a local plugin, which means it is associated with a channel.
An instance can be opened for each channel.

**Usage**

Click and drag to define a region within the image that will be used to
calculate the histogram.  To take the histogram of the full image, click
the button in the UI labeled "Full Image".

.. note:: Depending on the size of the image, calculating the
          full histogram may take time.

If a new image is selected for the channel, the histogram plot will be
recalculated based on the current parameters with the new data.

Unless disabled in the settings file for the histogram plugin, a line of
simple statistics for the box is calculated and shown in a line below the
plot.

**UI Controls**

Three radio buttons at the bottom of the UI are used to control the
effects of the click/drag action:

* select "Move" to drag the region to a different location
* select "Draw" to draw a new region
* select "Edit" to edit the region

To make a log plot of the histogram, check the "Log Histogram" checkbox.
To plot by the full range of values in the image instead of by the range
within the cut values, uncheck the "Plot By Cuts" checkbox.

The "NumBins" parameter determines how many bins are used in calculating
the histogram.  Type a number in the box and press "Enter" to change the
default value.

**Cut Levels Convenience Controls**

Because a histogram is useful feedback for setting the cut levels,
controls are provided in the UI for setting the low and high cut levels
in the image, as well as for performing an auto cut levels, according to
the auto cut levels settings in the channel preferences.

**User Configuration**

"""
import numpy as np

from ginga.gw import Widgets
from ginga import GingaPlugin
from ginga import AutoCuts

try:
    from ginga.gw import Plot
    from ginga.util import plots
    have_mpl = True
except ImportError:
    have_mpl = False

__all__ = ['Histogram']


class Histogram(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Histogram, self).__init__(fv, fitsimage)

        self.layertag = 'histogram-canvas'
        self.histtag = None
        # If True, limits X axis to lo/hi cut levels
        self.xlimbycuts = True
        self._split_sizes = [400, 500]

        # get Histogram preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Histogram')
        self.settings.add_defaults(draw_then_move=True, num_bins=2048,
                                   hist_color='aquamarine', show_stats=True,
                                   maxdigits=7)
        self.settings.load(onError='silent')

        # Set up histogram control parameters
        self.histcolor = self.settings.get('hist_color', 'aquamarine')
        self.numbins = self.settings.get('num_bins', 2048)
        self.autocuts = AutoCuts.Histogram(self.logger)

        # for formatting statistics line
        self.show_stats = self.settings.get('show_stats', True)
        maxdigits = self.settings.get('maxdigits', 7)
        self.fmt_cell = '{:< %d.%dg}' % (maxdigits - 1, maxdigits // 2)

        self.dc = self.fv.get_draw_classes()

        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(True)
        canvas.set_drawtype('rectangle', color='cyan', linestyle='dash',
                            drawdims=True)
        canvas.set_callback('draw-event', self.draw_cb)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.add_draw_mode('move', down=self.drag,
                             move=self.drag, up=self.update)
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.set_surface(self.fitsimage)
        canvas.set_draw_mode('draw')
        self.canvas = canvas

        fitssettings = fitsimage.get_settings()
        for name in ['cuts']:
            fitssettings.get_setting(name).add_callback(
                'set', self.cutset_ext_cb, fitsimage)
        self.gui_up = False

    def build_gui(self, container):
        if not have_mpl:
            raise ImportError('Install matplotlib to use this plugin')

        top = Widgets.VBox()
        top.set_border_width(4)

        # Make the cuts plot
        box, sw, orientation = Widgets.get_oriented_box(container)
        box.set_border_width(4)
        box.set_spacing(2)

        paned = Widgets.Splitter(orientation=orientation)
        self.w.splitter = paned

        self.plot = plots.Plot(logger=self.logger,
                               width=400, height=400)
        ax = self.plot.add_axis()
        ax.grid(True)
        w = Plot.PlotWidget(self.plot)
        w.resize(400, 400)
        paned.add_widget(Widgets.hadjust(w, orientation))

        vbox = Widgets.VBox()
        vbox.set_border_width(2)

        # for statistics line
        self.w.stats1 = Widgets.Label('')
        vbox.add_widget(self.w.stats1)

        captions = (('Cut Low:', 'label', 'Cut Low', 'entry'),
                    ('Cut High:', 'label', 'Cut High', 'entry',
                     'Cut Levels', 'button'),
                    ('Auto Levels', 'button'),
                    ('Log Histogram', 'checkbutton',
                     'Plot By Cuts', 'checkbutton'),
                    ('NumBins:', 'label', 'NumBins', 'entry'),
                    ('Full Image', 'button'),
                    )

        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)
        b.cut_levels.set_tooltip("Set cut levels manually")
        b.auto_levels.set_tooltip("Set cut levels by algorithm")
        b.cut_low.set_tooltip("Set low cut level (press Enter)")
        b.cut_high.set_tooltip("Set high cut level (press Enter)")
        b.log_histogram.set_tooltip("Use the log of the pixel values for the "
                                    "histogram (empty bins map to 10^-1)")
        b.plot_by_cuts.set_tooltip("Only show the part of the histogram "
                                   "between the cuts")
        b.numbins.set_tooltip("Number of bins for the histogram")
        b.full_image.set_tooltip("Use the full image for calculating the "
                                 "histogram")
        b.numbins.set_text(str(self.numbins))
        b.cut_low.add_callback('activated', lambda w: self.cut_levels())
        b.cut_high.add_callback('activated', lambda w: self.cut_levels())
        b.cut_levels.add_callback('activated', lambda w: self.cut_levels())
        b.auto_levels.add_callback('activated', lambda w: self.auto_levels())

        b.log_histogram.set_state(self.plot.logy)
        b.log_histogram.add_callback('activated', self.log_histogram_cb)
        b.plot_by_cuts.set_state(self.xlimbycuts)
        b.plot_by_cuts.add_callback('activated', self.plot_by_cuts_cb)
        b.numbins.add_callback('activated', lambda w: self.set_numbins_cb())
        b.full_image.add_callback('activated', lambda w: self.full_image_cb())

        fr = Widgets.Frame("Histogram")
        vbox.add_widget(w)
        fr.set_widget(vbox)
        box.add_widget(fr, stretch=0)
        paned.add_widget(sw)
        paned.set_sizes(self._split_sizes)

        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        btn1 = Widgets.RadioButton("Move")
        btn1.set_state(mode == 'move')
        btn1.add_callback('activated',
                          lambda w, val: self.set_mode_cb('move', val))
        btn1.set_tooltip("Choose this to position box")
        self.w.btn_move = btn1
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Draw", group=btn1)
        btn2.set_state(mode == 'draw')
        btn2.add_callback('activated',
                          lambda w, val: self.set_mode_cb('draw', val))
        btn2.set_tooltip("Choose this to draw a replacement box")
        self.w.btn_draw = btn2
        hbox.add_widget(btn2)

        btn3 = Widgets.RadioButton("Edit", group=btn1)
        btn3.set_state(mode == 'edit')
        btn3.add_callback('activated',
                          lambda w, val: self.set_mode_cb('edit', val))
        btn3.set_tooltip("Choose this to edit a box")
        self.w.btn_edit = btn3
        hbox.add_widget(btn3)

        if self.histtag is None:
            self.w.btn_move.set_enabled(False)
            self.w.btn_edit.set_enabled(False)

        hbox.add_widget(Widgets.Label(''), stretch=1)

        top.add_widget(paned, stretch=5)
        top.add_widget(hbox, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)
        self.gui_up = True

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        self.plot.set_titles(rtitle="Histogram")

        # insert canvas, if not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)

        except KeyError:
            # Add ruler layer
            p_canvas.add(self.canvas, tag=self.layertag)

        #self.canvas.delete_all_objects()
        self.resume()

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        # turn off any mode user may be in
        self.modes_off()

        self.canvas.ui_set_active(True)
        self.fv.show_status("Draw a rectangle with the right mouse button")

    def stop(self):
        self.gui_up = False
        self._split_sizes = self.w.splitter.get_sizes()
        # remove the rect from the canvas
        ## try:
        ##     self.canvas.delete_object_by_tag(self.histtag)
        ## except Exception:
        ##     pass
        ##self.histtag = None

        # remove the canvas from the image
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        self.fv.show_status("")

    def full_image_cb(self):
        canvas = self.canvas
        try:
            canvas.delete_object_by_tag(self.histtag)
        except Exception:
            pass

        image = self.fitsimage.get_image()
        width, height = image.get_size()
        x1, y1, x2, y2 = 0, 0, width - 1, height - 1
        tag = canvas.add(self.dc.Rectangle(x1, y1, x2, y2,
                                           color='cyan',
                                           linestyle='dash'))
        self.draw_cb(canvas, tag)

    def get_data(self, image, x1, y1, x2, y2, z=None):
        if z is not None:
            data = image.get_data()
            data = data[y1:y2, x1:x2, z]
        else:
            tup = image.cutout_adjust(x1, y1, x2, y2)
            data = tup[0]

        return data

    def histogram(self, image, x1, y1, x2, y2, z=None, pct=1.0, numbins=2048):
        self.logger.warning("This call will be deprecated soon. "
                            "Use get_data() and histogram_data().")
        data_np = self.get_data(image, x1, y1, x2, y2, z=z)
        return self.histogram_data(data_np, pct=pct, numbins=numbins)

    def histogram_data(self, data, pct=1.0, numbins=2048):
        return self.autocuts.calc_histogram(data, pct=pct, numbins=numbins)

    def redo(self):
        if self.histtag is None:
            return

        obj = self.canvas.get_object_by_tag(self.histtag)
        if obj.kind != 'compound':
            return True
        bbox = obj.objects[0]

        # Do histogram on the points within the rect
        image = self.fitsimage.get_image()
        self.plot.clear()

        numbins = self.numbins

        depth = image.get_depth()
        if depth != 3:
            data_np = self.get_data(image, int(bbox.x1), int(bbox.y1),
                                    int(bbox.x2), int(bbox.y2))
            res = self.histogram_data(data_np, pct=1.0, numbins=numbins)
            # used with 'steps-post' drawstyle, this x and y assignment
            # gives correct histogram-steps
            x = res.bins
            y = np.append(res.dist, res.dist[-1])
            ymax = y.max()
            if self.plot.logy:
                y = np.choose(y > 0, (.1, y))
            self.plot.plot(x, y, xtitle="Pixel value", ytitle="Number",
                           title="Pixel Value Distribution",
                           color='blue', alpha=1.0, drawstyle='steps-post')
        else:
            colors = ('red', 'green', 'blue')
            ymax = 0
            for z in range(depth):
                data_np = self.get_data(image, int(bbox.x1), int(bbox.y1),
                                        int(bbox.x2), int(bbox.y2), z=z)
                res = self.histogram_data(data_np, pct=1.0, numbins=numbins)
                # used with 'steps-post' drawstyle, this x and y assignment
                # gives correct histogram-steps
                x = res.bins
                y = np.append(res.dist, res.dist[-1])
                ymax = max(ymax, y.max())
                if self.plot.logy:
                    y = np.choose(y > 0, (.1, y))
                self.plot.plot(x, y, xtitle="Pixel value", ytitle="Number",
                               title="Pixel Value Distribution",
                               color=colors[z], alpha=0.33,
                               drawstyle='steps-post')

        # show cut levels
        loval, hival = self.fitsimage.get_cut_levels()
        self.loline = self.plot.ax.axvline(loval, 0.0, 0.99,
                                           linestyle='-', color='red')
        self.hiline = self.plot.ax.axvline(hival, 0.0, 0.99,
                                           linestyle='-', color='green')
        if self.xlimbycuts:
            self.plot.ax.set_xlim(loval, hival)

        # Make x axis labels a little more readable
        ## lbls = self.plot.ax.xaxis.get_ticklabels()
        ## for lbl in lbls:
        ##     lbl.set(rotation=45, horizontalalignment='right')

        self.w.cut_low.set_text(str(loval))
        self.w.cut_high.set_text(str(hival))
        self.plot.fig.canvas.draw()

        if self.show_stats:
            # calculate statistics on finite elements in box
            i = np.isfinite(data_np)
            if np.any(i):
                maxval = np.max(data_np[i])
                minval = np.min(data_np[i])
                meanval = np.mean(data_np[i])
                rmsval = np.sqrt(np.mean(np.square(data_np[i])))
                fmt_stat = "  Min: %s  Max: %s  Mean: %s  Rms: %s" % (
                    self.fmt_cell, self.fmt_cell, self.fmt_cell, self.fmt_cell)
                sum_text = fmt_stat.format(minval, maxval, meanval, rmsval)
            else:
                sum_text = "No finite data elements in cutout"
            self.w.stats1.set_text(sum_text)

        self.fv.show_status("Click or drag left mouse button to move region")
        return True

    def update(self, canvas, event, data_x, data_y, viewer):

        obj = self.canvas.get_object_by_tag(self.histtag)
        if obj.kind == 'compound':
            bbox = obj.objects[0]
        elif obj.kind == 'rectangle':
            bbox = obj
        else:
            return True

        # calculate center of bbox
        wd = bbox.x2 - bbox.x1
        dw = wd // 2
        ht = bbox.y2 - bbox.y1
        dh = ht // 2
        x, y = bbox.x1 + dw, bbox.y1 + dh

        # calculate offsets of move
        dx = (data_x - x)
        dy = (data_y - y)

        # calculate new coords
        x1, y1, x2, y2 = bbox.x1 + dx, bbox.y1 + dy, bbox.x2 + dx, bbox.y2 + dy

        try:
            canvas.delete_object_by_tag(self.histtag)
        except Exception:
            pass

        tag = canvas.add(self.dc.Rectangle(
            x1, y1, x2, y2, color='cyan', linestyle='dash'))

        self.draw_cb(canvas, tag)
        return True

    def drag(self, canvas, event, data_x, data_y, viewer):

        obj = self.canvas.get_object_by_tag(self.histtag)
        if obj.kind == 'compound':
            bbox = obj.objects[0]
        elif obj.kind == 'rectangle':
            bbox = obj
        else:
            return True

        # calculate center of bbox
        wd = bbox.x2 - bbox.x1
        dw = wd // 2
        ht = bbox.y2 - bbox.y1
        dh = ht // 2
        x, y = bbox.x1 + dw, bbox.y1 + dh

        # calculate offsets of move
        dx = (data_x - x)
        dy = (data_y - y)

        # calculate new coords
        x1, y1, x2, y2 = bbox.x1 + dx, bbox.y1 + dy, bbox.x2 + dx, bbox.y2 + dy

        if obj.kind == 'compound':
            try:
                canvas.delete_object_by_tag(self.histtag)
            except Exception:
                pass

            self.histtag = canvas.add(self.dc.Rectangle(
                x1, y1, x2, y2, color='cyan', linestyle='dash'))
        else:
            bbox.x1, bbox.y1, bbox.x2, bbox.y2 = x1, y1, x2, y2
            canvas.redraw(whence=3)

        return True

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        if obj.kind != 'rectangle':
            return True
        canvas.delete_object_by_tag(tag)

        if self.histtag:
            try:
                canvas.delete_object_by_tag(self.histtag)
            except Exception:
                pass

        x1, y1, x2, y2 = obj.get_llur()

        tag = canvas.add(self.dc.CompoundObject(
            self.dc.Rectangle(x1, y1, x2, y2,
                              color=self.histcolor),
            self.dc.Text(x1, y2, "Histogram",
                         color=self.histcolor)))
        self.histtag = tag

        self.w.btn_move.set_enabled(True)
        self.w.btn_edit.set_enabled(True)

        move_flag = self.settings.get('draw_then_move', True)
        if move_flag:
            self.set_mode('move')

        return self.redo()

    def edit_cb(self, canvas, obj):
        if obj.kind != 'rectangle':
            return True

        # Get the compound object that sits on the canvas.
        # Make sure edited rectangle was our histogram rectangle.
        c_obj = self.canvas.get_object_by_tag(self.histtag)
        if ((c_obj.kind != 'compound') or (len(c_obj.objects) < 2) or
                (c_obj.objects[0] != obj)):
            return False

        # reposition other elements to match
        x1, y1, x2, y2 = obj.get_llur()
        text = c_obj.objects[1]
        text.x, text.y = x1, y2 + 4
        self.fitsimage.redraw(whence=3)

        return self.redo()

    def cut_levels(self):
        reslvls = None

        try:
            loval = float(self.w.cut_low.get_text())
            hival = float(self.w.cut_high.get_text())

            reslvls = self.fitsimage.cut_levels(loval, hival)

        except Exception as e:
            errmsg = 'Error cutting levels: {0}'.format(str(e))
            self.fv.show_status(errmsg)
            self.logger.error(errmsg)

        else:
            if self.xlimbycuts:
                self.redo()

        return reslvls

    def auto_levels(self):
        self.fitsimage.auto_levels()

    def cutset_ext_cb(self, setting, value, fitsimage):
        if not self.gui_up:
            return

        t_ = fitsimage.get_settings()
        loval, hival = t_['cuts']

        try:
            self.loline.remove()
            self.hiline.remove()
        except Exception:
            pass
        self.loline = self.plot.ax.axvline(loval, 0.0, 0.99,
                                           linestyle='-', color='black')
        self.hiline = self.plot.ax.axvline(hival, 0.0, 0.99,
                                           linestyle='-', color='black')
        self.w.cut_low.set_text(str(loval))
        self.w.cut_high.set_text(str(hival))
        #self.plot.fig.canvas.draw()
        self.redo()

    def set_numbins_cb(self):
        self.numbins = int(self.w.numbins.get_text())
        self.redo()

    def log_histogram_cb(self, w, val):
        self.plot.logy = val
        if (self.histtag is not None) and self.gui_up:
            # self.histtag is None means no data is loaded yet
            self.redo()

    def plot_by_cuts_cb(self, w, val):
        self.xlimbycuts = val
        if (self.histtag is not None) and self.gui_up:
            # self.histtag is None means no data is loaded yet
            self.redo()

    def edit_select_box(self):
        if self.histtag is not None:
            obj = self.canvas.get_object_by_tag(self.histtag)
            if obj.kind != 'compound':
                return True
            # drill down to reference shape
            bbox = obj.objects[0]
            self.canvas.edit_select(bbox)
        else:
            self.canvas.clear_selected()
        self.canvas.update_canvas()

    def set_mode_cb(self, mode, tf):
        """Called when one of the Move/Draw/Edit radio buttons is selected."""
        if tf:
            self.canvas.set_draw_mode(mode)
            if mode == 'edit':
                self.edit_select_box()
        return True

    def set_mode(self, mode):
        self.canvas.set_draw_mode(mode)
        self.w.btn_move.set_state(mode == 'move')
        self.w.btn_draw.set_state(mode == 'draw')
        self.w.btn_edit.set_state(mode == 'edit')

    def __str__(self):
        return 'histogram'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Histogram', package='ginga')

# END



#
# RemoteImage.py -- Mixin for images that live on a remote server
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

from ginga.AstroImage import AstroImage, AstroHeader

class RemoteImage(AstroImage):

    def __init__(self, proxy, metadata=None, logger=None,
                 #wcsclass=wcsClass, ioclass=ioClass,
                 inherit_primary_header=False):

        self._proxy = proxy
        self._shape = tuple([])
        self.id = None

        AstroImage.__init__(self, data_np=None, metadata=metadata,
                            logger=logger, #wcsclass=wcsClass, ioclass=ioClass,
                            inherit_primary_header=inherit_primary_header)
        self._data = None

    @property
    def shape(self):
        return self._shape

    def load_file(self, filepath, numhdu=None, naxispath=None):
        self.id = filepath
        shape, header = self._proxy.load_file(filepath, numhdu=numhdu,
                                              naxispath=naxispath)
        self._shape = shape

        ahdr = { key: header[key]['value'] for key in header.keys() }
        # this will set our local wcs
        self.update_keywords(ahdr)

    def _set_minmax(self, noinf=False):
        self.minval, self.maxval = self._proxy.get_minmax(self.id, noinf=False)
        (self.minval_noinf,
         self.maxval_noinf) = self._proxy.get_minmax(self.id, noinf=True)

    def _get_data(self):
        if self._data is None:
            self._data = self._proxy.get_data(self.id)
        return self._data

    def _slice(self, view):
        """
        Send view to remote server and do slicing there.
        """
        if self._data is not None:
            return self._data[view]
        return self._proxy.get_view(self.id, view)

    def get_data_xy(self, x, y):
        # TODO: cache some points for faster response?
        return self._proxy.get_data_xy(self.id, x, y)

    def get_pixels_on_line(self, x1, y1, x2, y2):
        return self._proxy.get_pixels_on_line(self.id, x1, y1, x2, y2)



import numpy as np
import pytest

from ginga.util import iqcalc


@pytest.mark.parametrize(
    ('arr', 'ans'),
    [(np.arange(5), 2),
     (np.array([1, np.inf, 3, np.nan, 5]), 3),
     (np.arange(10).reshape(2, 5), 4.5)])
def test_get_mean_median(arr, ans):
    assert iqcalc.get_mean(arr) == ans
    assert iqcalc.get_median(arr) == ans


def test_get_mean_median_nan():
    arr = np.array([np.nan, np.inf])
    assert np.isnan(iqcalc.get_mean(arr))
    assert np.isnan(iqcalc.get_median(arr))



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
``WCSMatch`` is a global plugin for the Ginga image viewer that allows
you to roughly align images with different scales and orientations
using the images' World Coordinate System (WCS) for viewing purposes.

**Plugin Type: Global**

``WCSMatch`` is a global plugin.  Only one instance can be opened.

**Usage**

To use, simply start the plugin, and from the plugin GUI select a
channel from the drop-down menu labeled "Reference Channel".  The
image contained in that channel will be used as a reference for
synchronizing the images in the other channels.

The channels will be synchronized in viewing (pan, scale (zoom),
transforms (flips) and rotation.  The checkboxes "Match Pan",
"Match Scale", "Match Transforms" and "Match Rotation" can be
checked or not to control which attributes are synchronized between
channels.

To completely "unlock" the synchronization, simply select "None"
from the "Reference Channel" drop-down menu.

Currently, there is no way to limit the channels that are affected
by the plugin.

"""
from ginga import GingaPlugin
from ginga.gw import Widgets
from ginga.util import wcs
from ginga.misc import Bunch

__all__ = ['WCSMatch']


class WCSMatch(GingaPlugin.GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super(WCSMatch, self).__init__(fv)

        self.chnames = []
        self.ref_channel = None
        self.ref_image = None
        self.gui_up = False
        self._cur_opn_viewer = None
        self._match = dict(pan=True, scale=True, transforms=True,
                           rotation=True)

        fv.add_callback('add-channel', self.add_channel)
        fv.add_callback('delete-channel', self.delete_channel)

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        vbox, sw, orientation = Widgets.get_oriented_box(container)
        vbox.set_border_width(4)
        vbox.set_spacing(2)

        fr = Widgets.Frame("WCS Match")

        captions = (('Reference Channel:', 'label',
                     'ref channel', 'combobox'),
                    ('Match Pan', 'checkbutton',
                     'Match Scale', 'checkbutton'),
                    ('Match Transforms', 'checkbutton',
                    'Match Rotation', 'checkbutton'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w = b

        b.ref_channel.add_callback('activated', self._set_reference_channel_cb)
        self.w.match_pan.set_state(self._match['pan'])
        self.w.match_pan.set_tooltip("Match pan position of reference image")
        self.w.match_pan.add_callback('activated',
                                      self.set_match_cb, 'pan')
        self.w.match_scale.set_state(self._match['scale'])
        self.w.match_scale.add_callback('activated',
                                        self.set_match_cb, 'scale')
        self.w.match_scale.set_tooltip("Match scale of reference image")
        self.w.match_transforms.set_state(self._match['transforms'])
        self.w.match_transforms.add_callback('activated',
                                             self.set_match_cb, 'transforms')
        self.w.match_transforms.set_tooltip("Match transforms of reference image")
        self.w.match_rotation.set_state(self._match['rotation'])
        self.w.match_rotation.add_callback('activated',
                                           self.set_match_cb, 'rotation')
        self.w.match_rotation.set_tooltip("Match rotation of reference image")

        fr.set_widget(w)
        vbox.add_widget(fr, stretch=0)

        spacer = Widgets.Label('')
        vbox.add_widget(spacer, stretch=1)

        top.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)

        self.gui_up = True
        self._reset_channels_gui()

    def add_channel(self, viewer, channel):
        info = Bunch.Bunch(chinfo=channel)
        channel.extdata._wcsmatch_info = info

        # Add callbacks to the viewer for all the scale, pan, rotation and
        # transform settings changes
        chviewer = channel.fitsimage
        fitssettings = chviewer.get_settings()
        fitssettings.get_setting('scale').add_callback(
            'set', self.zoomset_cb, chviewer, info)
        fitssettings.get_setting('rot_deg').add_callback(
            'set', self.rotset_cb, chviewer, info)
        for name in ('flip_x', 'flip_y', 'swap_xy'):
            fitssettings.get_setting(name).add_callback(
                'set', self.xfmset_cb, chviewer, info)
        fitssettings.get_setting('pan').add_callback(
            'set', self.panset_cb, chviewer, info)
        self.fv.gui_do(self._reset_channels_gui)

    def delete_channel(self, viewer, channel):
        chname = channel.name
        self.logger.debug("deleting channel %s" % (chname))
        self.active = None
        self.info = None
        self.fv.gui_do(self._reset_channels_gui)

    def _reset_channels_gui(self):
        self.chnames = list(self.fv.get_channel_names())
        self.chnames.sort()
        self.chnames.insert(0, "None")
        if not self.gui_up:
            return
        self.w.ref_channel.clear()
        for chname in self.chnames:
            self.w.ref_channel.append_text(chname)

    # CALLBACKS

    def _update_all(self):
        if self.ref_channel is None:
            return
        chinfo = self.ref_channel
        chviewer = chinfo.fitsimage
        self.zoomset(chviewer, chinfo)
        self.rotset(chviewer, chinfo)
        self.xfmset(chviewer, chinfo)
        self.panset(chviewer, chinfo)

    def _set_reference_channel(self, chname):
        if chname == 'None':
            chname = None
        if chname is None:
            self.ref_image = None
            self.ref_channel = None
            self.logger.info("turning off channel synchronization")
            return

        chinfo = self.fv.get_channel(chname)
        self.ref_channel = chinfo
        chviewer = chinfo.fitsimage
        self.ref_image = chviewer.get_image()

        # reset the scale base to be identical in both axes for the
        # reference image
        chviewer.set_scale_base_xy(1.0, 1.0)

        self._update_all()

        self.logger.info("set reference channel to '%s'" % (chname))

    def _set_reference_channel_cb(self, w, idx):
        """This is the GUI callback for the control that sets the reference
        channel.
        """
        chname = self.chnames[idx]
        self._set_reference_channel(chname)

    def set_reference_channel(self, chname):
        """This is the API call to set the reference channel.
        """
        # change the GUI control to match
        idx = self.chnames.index(str(chname))
        self.w.ref_channel.set_index(idx)
        return self._set_reference_channel(chname)

    def set_match_cb(self, w, tf, key):
        # remember, in case we are closed and reopened
        self._match.update(dict(pan=self.w.match_pan.get_state(),
                                scale=self.w.match_scale.get_state(),
                                transforms=self.w.match_transforms.get_state(),
                                rotation=self.w.match_rotation.get_state()))

        if key == 'scale' and tf is False:
            self.reset_scale_skew()

        self._update_all()

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def start(self):
        pass

    def stop(self):
        self.ref_channel = None
        self.ref_image = None
        self.fv.show_status("")

    def get_other_channels(self, myname):
        return set(self.fv.get_channel_names()) - set([myname])

    def zoomset_cb(self, setting, value, chviewer, info):
        """This callback is called when a channel window is zoomed.
        """
        return self.zoomset(chviewer, info.chinfo)

    def zoomset(self, chviewer, chinfo):
        # Don't do anything if we are not active
        if not self.gui_up or self.ref_image is None:
            return

        if self._cur_opn_viewer is not None:
            return
        if not self._match['scale']:
            return

        self._cur_opn_viewer = chviewer
        try:
            self.scale_all_relative(chviewer, chinfo)

        finally:
            self._cur_opn_viewer = None

    def scale_all_relative(self, chviewer, chinfo):
        if self.ref_image is None:
            return

        # get native scale relative to reference image
        image = chviewer.get_image()
        ort = wcs.get_relative_orientation(image, self.ref_image)
        self.logger.info("scale for channel '%s' relative to ref image "
                         "%f,%f" % (chinfo.name, ort.rscale_x, ort.rscale_y))

        scale_x, scale_y = chviewer.get_scale_xy()

        chg_x, chg_y = scale_x / ort.rscale_x, scale_y / ort.rscale_y
        self.logger.info("scale changed for channel '%s' by %f,%f" % (
            chinfo.name, chg_x, chg_y))

        # for all other channels except ours
        chnames = self.get_other_channels(chinfo.name)
        for chname in chnames:
            chinfo2 = self.fv.get_channel(chname)

            # calculate scale from orientation to reference image
            image = chinfo2.fitsimage.get_image()
            if image is None:
                continue
            ort = wcs.get_relative_orientation(image, self.ref_image)

            new_scale_x, new_scale_y = (ort.rscale_x * chg_x,
                                        ort.rscale_y * chg_y)
            # apply that scale
            self.logger.info("changing scale for channel '%s' to %f,%f" % (
                chinfo2.name, new_scale_x, new_scale_y))
            chinfo2.fitsimage.scale_to(new_scale_x, new_scale_y)

    def reset_scale_skew(self):
        """Set the X/Y scaling factors to be equal in each channel.
        """
        chnames = self.fv.get_channel_names()
        for chname in chnames:
            chinfo = self.fv.get_channel(chname)

            scales = chinfo.fitsimage.get_scale_xy()
            fac = min(*scales)
            chinfo.fitsimage.scale_to(fac, fac)

    def rotset_cb(self, setting, value, chviewer, info):
        """This callback is called when a channel window is rotated.
        """
        return self.rotset(chviewer, info.chinfo)

    def rotset(self, chviewer, chinfo):
        # Don't do anything if we are not active
        if not self.gui_up or self.ref_image is None:
            return

        if self._cur_opn_viewer is not None:
            return
        if not self._match['rotation']:
            return

        self._cur_opn_viewer = chviewer
        try:
            self.rotate_all_relative(chviewer, chinfo)

        finally:
            self._cur_opn_viewer = None

    def rotate_all_relative(self, chviewer, chinfo):
        if self.ref_image is None:
            return

        # get native scale relative to reference image
        image = chviewer.get_image()
        if self.ref_image is None:
            return
        ort = wcs.get_relative_orientation(image, self.ref_image)
        self.logger.info("rotation for channel '%s' relative to ref image "
                         "%f" % (chinfo.name, ort.rrot_deg))

        rot_deg = chviewer.get_rotation()

        chg_rot_deg = rot_deg + ort.rrot_deg
        self.logger.info("rotation changed for channel '%s' by %f" % (
            chinfo.name, chg_rot_deg))

        # for all other channels except ours
        chnames = self.get_other_channels(chinfo.name)
        for chname in chnames:
            chinfo2 = self.fv.get_channel(chname)

            # Get relative rotation of their image
            image = chinfo2.fitsimage.get_image()
            if image is None:
                continue
            ort = wcs.get_relative_orientation(image, self.ref_image)

            # Apply that rotation
            new_rot_deg = ort.rrot_deg + chg_rot_deg
            self.logger.info("changing rot for channel '%s' to %f" % (
                chinfo2.name, new_rot_deg))
            chinfo2.fitsimage.rotate(new_rot_deg)

    def panset_cb(self, setting, value, chviewer, info):
        """This callback is called when a channel window is panned.
        """
        return self.panset(chviewer, info.chinfo)

    def panset(self, chviewer, chinfo):
        # Don't do anything if we are not active
        if not self.gui_up or self.ref_image is None:
            return

        if self._cur_opn_viewer is not None:
            return
        if not self._match['pan']:
            return

        self._cur_opn_viewer = chviewer
        try:
            self.pan_all_relative(chviewer, chinfo)

        finally:
            self._cur_opn_viewer = None

    def pan_all_relative(self, chviewer, chinfo):
        if self.ref_image is None:
            return

        image = chviewer.get_image()
        if self.ref_image is None:
            return

        pan_ra, pan_dec = chviewer.get_pan(coord='wcs')

        # for all other channels except ours
        chnames = self.get_other_channels(chinfo.name)
        for chname in chnames:
            chinfo2 = self.fv.get_channel(chname)

            # set pan position on their viewer
            image = chinfo2.fitsimage.get_image()
            if image is None:
                continue
            data_x, data_y = image.radectopix(pan_ra, pan_dec)
            chinfo2.fitsimage.panset_xy(data_x, data_y)

    def xfmset_cb(self, setting, value, chviewer, info):
        """This callback is called when a channel window is transformed
        (flipped, or swap axes).
        """
        return self.xfmset(chviewer, info.chinfo)

    def xfmset(self, chviewer, chinfo):
        # Don't do anything if we are not active
        if not self.gui_up or self.ref_image is None:
            return

        if self._cur_opn_viewer is not None:
            return
        if not self._match['transforms']:
            return

        self._cur_opn_viewer = chviewer
        try:
            self.transform_all_relative(chviewer, chinfo)

        finally:
            self._cur_opn_viewer = None

    def transform_all_relative(self, chviewer, chinfo):
        if self.ref_image is None:
            return

        image = chviewer.get_image()
        if self.ref_image is None:
            return

        flip_x, flip_y, swap_xy = chviewer.get_transforms()

        # for all other channels except ours
        chnames = self.get_other_channels(chinfo.name)
        for chname in chnames:
            chinfo2 = self.fv.get_channel(chname)

            # set our pan position on their viewer
            image = chinfo2.fitsimage.get_image()
            if image is None:
                continue
            chinfo2.fitsimage.transform(flip_x, flip_y, swap_xy)

    def __str__(self):
        return 'wcsmatch'

# END



#
# MplHelp.py -- help classes for Matplotlib drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

from ginga import colors
from ginga.misc import Bunch, Callback
from ginga.fonts import font_asst

import matplotlib.textpath as textpath


class Pen(object):
    def __init__(self, color='black', linewidth=1, linestyle='solid'):
        self.color = color
        self.linewidth = linewidth
        if linestyle == 'dash':
            linestyle = 'dashdot'
        self.linestyle = linestyle


class Brush(object):
    def __init__(self, color='black', fill=False):
        self.color = color
        self.fill = fill


class Font(object):
    def __init__(self, fontname='sans', fontsize=12.0, color='black'):
        fontname = font_asst.resolve_alias(fontname, fontname)
        self.fontname = fontname
        self.fontsize = fontsize
        self.color = color

    def get_fontdict(self):
        fontdict = dict(color=self.color, family=self.fontname,
                        size=self.fontsize, transform=None)
        return fontdict


def load_font(font_name, font_file):
    # TODO!
    ## raise ValueError("Loading fonts dynamically is an unimplemented"
    ##                  " feature for matplotlib back end")
    return font_name


class MplContext(object):

    def __init__(self, axes):
        self.axes = axes
        self.kwdargs = dict()
        self.stack = []

    def set_canvas(self, axes):
        self.axes = axes

    def init(self, **kwdargs):
        self.kwdargs = dict()
        self.kwdargs.update(kwdargs)

    def set(self, **kwdargs):
        self.kwdargs.update(kwdargs)

    def push(self, allow=[]):
        self.stack.append(self.kwdargs.copy())
        d = {name: self.kwdargs[name]
             for name in allow if name in self.kwdargs}
        self.kwdargs = d

    def pop(self):
        self.kwdargs = self.stack.pop()

    def update_fill(self, brush):
        if brush is None:
            self.kwdargs['fill'] = False
            return

        self.kwdargs['fill'] = True
        self.kwdargs['facecolor'] = brush.color

    def update_line(self, pen):
        self.kwdargs['color'] = pen.color
        self.kwdargs['linewidth'] = pen.linewidth
        self.kwdargs['linestyle'] = pen.linestyle

    def update_patch(self, pen, brush):
        self.update_fill(brush)

        if self.kwdargs['fill']:
            line_color_attr = 'facecolor'
            if 'facecolor' in self.kwdargs:
                line_color_attr = 'edgecolor'
        else:
            line_color_attr = 'color'

        self.kwdargs[line_color_attr] = pen.color
        self.kwdargs['linewidth'] = pen.linewidth
        self.kwdargs['linestyle'] = pen.linestyle

    def get_color(self, color, alpha):
        if isinstance(color, str) or isinstance(color, type(u"")):
            r, g, b = colors.lookup_color(color)
        elif isinstance(color, tuple):
            # color is assumed to be a 3-tuple of RGBA values as floats
            # between 0 and 1
            r, g, b = color
        else:
            r, g, b = 1.0, 1.0, 1.0

        return (r, g, b, alpha)

    def get_pen(self, color, alpha=1.0, linewidth=1, linestyle='solid'):
        color = self.get_color(color, alpha=alpha)
        return Pen(color=color, linewidth=linewidth, linestyle=linestyle)

    def get_brush(self, color, alpha=1.0):
        color = self.get_color(color, alpha=alpha)
        return Brush(color=color, fill=True)

    def get_font(self, name, size, color, alpha=1.0):
        color = self.get_color(color, alpha=alpha)
        return Font(fontname=name, fontsize=size, color=color)

    def text_extents(self, text, font):
        # This is not completely accurate because it depends a lot
        # on the renderer used, but that is complicated under Mpl
        t = textpath.TextPath((0, 0), text, size=font.fontsize,
                              prop=font.fontname)
        bb = t.get_extents()
        wd, ht = bb.width, bb.height
        return (wd, ht)


class Timer(Callback.Callbacks):
    """Abstraction of a GUI-toolkit implemented timer."""

    def __init__(self, duration=0.0, mplcanvas=None):
        """Create a timer set to expire after `duration` sec.
        """
        super(Timer, self).__init__()

        self.duration = duration
        # For storing aritrary data with timers
        self.data = Bunch.Bunch()

        self._timer = mplcanvas.new_timer()
        self._timer.single_shot = True
        self._timer.add_callback(self._redirect_cb)

        for name in ('expired', 'canceled'):
            self.enable_callback(name)

    def start(self, duration=None):
        """Start the timer.  If `duration` is not None, it should
        specify the time to expiration in seconds.
        """
        if duration is None:
            duration = self.duration

        self.set(duration)

    def set(self, duration):

        self.stop()

        # Matplotlib timer set in milliseconds
        time_ms = int(duration * 1000.0)
        self._timer.interval = time_ms
        self._timer.start()

    def _redirect_cb(self):
        self.make_callback('expired')

    def stop(self):
        try:
            self._timer.stop()
        except Exception:
            pass

    def cancel(self):
        """Cancel this timer.  If the timer is not running, there
        is no error.
        """
        self.stop()
        self.make_callback('canceled')

    clear = cancel

# END



#
# Imexam.py -- Imexam plugin for Ginga reference viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import inspect
import traceback
import logging
from io import StringIO

from ginga import GingaPlugin
from ginga.gw import Widgets
from ginga.gw import Plot
from ginga.util import plots
from ginga.misc import Bunch

have_imexam = False
try:
    from imexam.imexamine import Imexamine
    have_imexam = True

except ImportError:
    pass


class Imexam(GingaPlugin.LocalPlugin):
    """
    This is an experimental Ginga plugin for the "imexam" package.
    To use it you will need to install the "imexam" package:

        https://github.com/spacetelescope/imexam

    To install this plugin:
      $ mkdir $HOME/.ginga/plugins
      $ cp Imexam.py $HOME/.ginga/plugins/.

    To use:
      $ ginga ... --plugins=Imexam

      Then from the "Operations" menu, choose "Imexam".

    KNOWN ISSUES:
    - You need ginga v2.6.0.dev

    - When a plot is created for the first time, it will force the
      focus away from the channel viewer.  This means that keystrokes
      will not be recognized in the viewer again until you give it the
      focus back (by say, clicking in the window)

    - It makes the most sense to use the plugin with the channels
      workspace in "MDI" mode, although it works fine in other
      configurations.

    - Closing the plot windows is only possible currently by making
      sure the window has the focus and then using the workspace toolbar
      "-" button to delete it.

    - If you close an active plot window, you will need to press the
      "Detach Plot" button before plotting will work again--it doesn't
      recognize that the window has been closed.

    """

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(Imexam, self).__init__(fv, fitsimage)

        # get Imexam preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.createCategory('plugin_Imexam')
        self.settings.addDefaults(font='Courier', fontsize=12,
                                  plots_in_workspace=False)
        self.settings.load(onError='silent')

        self.layertag = 'imexam-canvas'
        self.imexam_active = False

        # this is our imexamine object
        self.imex = Imexamine()

        # capture the stdout logger from imexam
        self.log_capture_string = StringIO()
        self.stream_handler = logging.StreamHandler(self.log_capture_string)
        self.stream_handler.setLevel(logging.INFO)
        self.imex.log.addHandler(self.stream_handler)

        self.dc = fv.get_draw_classes()
        canvas = self.dc.DrawingCanvas()
        canvas.set_callback('key-press', self.key_press_cb)
        canvas.set_surface(self.fitsimage)
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.enable_draw(False)
        canvas.name = 'Imexam-canvas'
        self.canvas = canvas

        self._plot = None
        self._plot_w = None
        self._plot_idx = 0
        self._plots_in_ws = self.settings.get('plots_in_workspace', False)
        self.w = Bunch.Bunch()

    def build_gui(self, container):
        if not have_imexam:
            raise Exception("Please install 'imexam' to use this plugin")

        top = Widgets.VBox()
        top.set_border_width(4)

        fontsize = self.settings.get('fontsize', 12)

        msg_font = self.fv.get_font('sans', fontsize)
        tw = Widgets.TextArea(wrap=False, editable=False)
        tw.set_font(msg_font)
        self.tw = tw

        fr = Widgets.Expander("Instructions")
        fr.set_widget(tw)
        top.add_widget(fr, stretch=0)

        fr = Widgets.Frame("Imexam output:")

        if not self._plots_in_ws:
            splitter = Widgets.Splitter(orientation='vertical')

            self.nb = Widgets.TabWidget()
            splitter.add_widget(self.nb)

        # this holds the messages returned from imexamine
        tw = Widgets.TextArea(wrap=False, editable=False)
        font = self.settings.get('font', 'Courier')
        fixed_font = self.fv.get_font(font, fontsize)
        tw.set_font(fixed_font)
        self.msg_res = tw

        if not self._plots_in_ws:
            splitter.add_widget(tw)
            fr.set_widget(splitter)
        else:
            fr.set_widget(tw)

        top.add_widget(fr, stretch=1)

        hbox = Widgets.HBox()
        btn = Widgets.Button('Detach Plot')
        btn.add_callback('activated', self.detach_plot_cb)
        btn.set_tooltip("Detach current plot and start a new one")
        hbox.add_widget(btn, stretch=0)
        btn = Widgets.Button('Clear Text')
        btn.add_callback('activated', self.clear_text_cb)
        btn.set_tooltip("Clear the imexam output")
        hbox.add_widget(btn, stretch=0)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(hbox, stretch=0)

        hbox = Widgets.HBox()
        lbl = Widgets.Label("Keys active:")
        hbox.add_widget(lbl)
        btn1 = Widgets.RadioButton("On")
        btn1.set_state(self.imexam_active)
        btn1.add_callback('activated', lambda w, val: self.set_active_cb(True, val))
        btn1.set_tooltip("Enable imexam keys")
        self.w.btn_on = btn1
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Off", group=btn1)
        btn2.set_state(not self.imexam_active)
        btn2.add_callback('activated', lambda w, val: self.set_active_cb(False, val))
        btn2.set_tooltip("Disable imexam keys")
        self.w.btn_off = btn2
        hbox.add_widget(btn2)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(hbox, stretch=0)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        self._plot = None
        self._plot_w = None
        self._plot_idx = 0
        self.make_new_figure()

        container.add_widget(top, stretch=1)

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def set_active(self, onoff, update_ui=False):
        self.imexam_active = onoff
        self.canvas.ui_setActive(onoff)

        if update_ui and 'btn_on' in self.w:
            if onoff:
                self.w.btn_on.set_state(True)
            else:
                self.w.btn_off.set_state(True)

        if onoff:
            msg = "Imexam keys are active"
        else:
            msg = "Imexam keys deactivated"

        self.fitsimage.onscreen_message(msg, delay=1.0)
        self.fv.show_status(msg)

    def set_active_cb(self, tf, onoff):
        if tf:
            self.set_active(onoff, update_ui=False)

    def detach_plot_cb(self, w):
        self._plot = None
        self._plot_w = None
        self.make_new_figure()

    def clear_text_cb(self, w):
        self.msg_res.clear()

    def instructions(self):
        lines = ["Key bindings:"]
        for key, tup in self.imex.imexam_option_funcs.items():
            func, descr = tup
            lines.append("  %s : %s" % (key, descr))

        text = '\n'.join(lines)
        self.tw.set_text(text)

    def start(self):
        self.instructions()
        p_canvas = self.fitsimage.get_canvas()
        if not p_canvas.has_object(self.canvas):
            p_canvas.add(self.canvas, tag=self.layertag)

        self.clear()
        self.resume()

    def pause(self):
        self.set_active(False, update_ui=True)

    def resume(self):
        self.set_active(True, update_ui=True)

    def stop(self):
        self.pause()

        # remove the canvas from the image
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        self.fv.show_status("")

    def redo(self):
        pass

    def make_new_figure(self):
        chname = self.fv.get_channel_name(self.fitsimage)

        wd, ht = 400, 300
        self._plot_idx += 1
        self._plot = plots.Plot(logger=self.logger,
                                width=wd, height=ht)

        name = "%s: Fig %d" % (chname, self._plot_idx)
        group = 10

        pw = Plot.PlotWidget(self._plot)

        vbox = Widgets.VBox()
        vbox.add_widget(pw, stretch=1)
        hbox = Widgets.HBox()
        hbox.add_widget(Widgets.Label(''), stretch=1)
        btn = Widgets.Button('Close Plot')
        btn.add_callback('activated', lambda w: self.close_plot(name, vbox))
        hbox.add_widget(btn, stretch=0)
        vbox.add_widget(hbox, stretch=0)

        # vbox.resize(wd, ht)
        self._plot_w = vbox

        if self._plots_in_ws:
            ws = self.fv.get_current_workspace()
            tab = self.fv.ds.add_tab(ws.name, vbox, group, name, name,
                                     data=dict(plot=self._plot))
        else:
            self.nb.add_widget(vbox, name)

        # imexam should get a clean figure
        fig = self._plot.get_figure()
        fig.clf()

    def close_plot(self, name, child):
        if child == self._plot_w:
            self.make_new_figure()

        if not self._plots_in_ws:
            self.nb.remove(child)

        return True

    def imexam_cmd(self, canvas, keyname, data_x, data_y, func):
        if not self.imexam_active:
            return False
        self.logger.debug("imexam_cb")

        # keyname = event.key
        self.logger.debug("key pressed: %s" % (keyname))

        image = self.fitsimage.get_image()
        if image is None:
            return False

        # inspect func to see what kind of things we can pass in
        args, varargs, varkw, defaults = inspect.getargspec(func)

        kwargs = dict()

        if 'data' in args:
            # pass the data array
            data_np = image.get_data()
            kwargs['data'] = data_np

        if 'fig' in args:
            # Make a new figure if we don't have one
            if self._plot is None:
                self.make_new_figure()
            kwargs['fig'] = self._plot.get_figure()
        self.log_capture_string.seek(0)
        self.log_capture_string.truncate()
        self.msg_res.append_text("----\ncmd: '%s'\n" % keyname)
        try:
            func(data_x, data_y, **kwargs)
            self.msg_res.append_text(self.log_capture_string.getvalue(),
                                     autoscroll=True)
        except Exception as e:
            self.msg_res.append_text(self.log_capture_string.getvalue(),
                                     autoscroll=True)
            # get any partial output
            errmsg = ("Error calling imexam function: %s" % (
                str(e)))
            self.msg_res.append_text(errmsg)

            # show traceback
            try:
                (type, value, tb) = sys.exc_info()
                tb_str = "\n".join(traceback.format_tb(tb))
            except Exception as e:
                tb_str = "Traceback information unavailable."

            self.msg_res.append_text(tb_str, autoscroll=True)

        return True

    def key_press_cb(self, canvas, keyname):

        # some keys that we explicitly can't handle from imexamine
        if keyname == '2':
            self.detach_plot_cb(None)
            return True

        try:
            # lookup imexamine function
            func, descr = self.imex.imexam_option_funcs[keyname]

        except KeyError:
            # no key binding for this in imexam
            return False

        data_x, data_y = self.fitsimage.get_last_data_xy()
        return self.imexam_cmd(self.canvas, keyname, data_x, data_y, func)


    def clear(self):
        self.canvas.delete_all_objects()
        return False

    def __str__(self):
        return 'imexam'

# END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
``Blink`` switches through the images shown in a channel at a rate
chosen by the user.  Alternatively, it can switch between channels
in the main workspace.  In both cases, the primary purpose is to
compare and contrast the images (within a channel, or across
channels) visually within a short timescale -- like blinking your
eyes.

**Plugin Type: Local or Global**

``Blink`` can be invoked either as a local plugin, in which case
it cycles through the images in the channel, or as a global
plugin, in which case it cycles through the channels.

Local plugins are started from the "Operations" button, while
global plugins are started from the "Plugins" menu.

**Usage**

Set the interval between image changes in terms of seconds in
the box labeled "Interval".  Then, press "Start Blink" to start
the timed cycling, and "Stop Blink" to stop the cycling.

You can change the number in "Interval" and press ``Enter`` to
dynamically change the cycle time while the cycle is running.

"""
import time

from ginga import GingaPlugin
from ginga.gw import Widgets

__all__ = ['Blink']


class Blink(GingaPlugin.LocalPlugin):

    def __init__(self, *args):
        # superclass defines some variables for us, like logger
        if len(args) == 2:
            super(Blink, self).__init__(*args)
        else:
            super(Blink, self).__init__(args[0], None)

        self.interval = 1.0
        self.blink_timer = self.fv.get_timer()
        self.blink_timer.set_callback('expired', self._blink_timer_cb)

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_Blink')
        self.settings.add_defaults(interval_max=30.0, interval_min=0.25)
        self.settings.load(onError='silent')

        # TODO: need to deprecate the blink_channels setting
        # the mode is now determined by whether the plugin is loaded
        # as a local or global
        self.blink_channels = (self.fitsimage is None)

        self.ival_min = self.settings.get('interval_min', 0.25)
        self.ival_max = self.settings.get('interval_max', 30.0)

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        vbox, sw, orientation = Widgets.get_oriented_box(container)
        vbox.set_border_width(4)
        vbox.set_spacing(2)

        fr = Widgets.Frame("Blink")
        vbox2 = Widgets.VBox()

        captions = (("Interval:", 'label', 'Interval', 'entry'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w = b

        b.interval.set_text(str(self.interval))
        b.interval.add_callback('activated', lambda w: self._set_interval_cb())
        b.interval.set_tooltip("Interval in seconds between changing images")
        vbox2.add_widget(w, stretch=0)

        captions = (("Start Blink", 'button', "Stop Blink", 'button'),
                    ("Max:", 'label', 'max', 'llabel',
                     "Min:", 'label', 'min', 'llabel'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)
        b.start_blink.add_callback('activated',
                                   lambda w: self._start_blink_cb())
        b.stop_blink.add_callback('activated',
                                  lambda w: self._stop_blink_cb())

        b.min.set_text(str(self.ival_min))
        b.max.set_text(str(self.ival_max))
        vbox2.add_widget(w, stretch=0)

        captions = (("Mode:", 'label', 'mode', 'llabel'),
                    )
        w, b = Widgets.build_info(captions, orientation=orientation)
        self.w.update(b)

        mode = 'blink channels'
        if not self.blink_channels:
            mode = 'blink images in channel'
        b.mode.set_text(mode)

        vbox2.add_widget(w, stretch=0)

        fr.set_widget(vbox2)
        vbox.add_widget(fr, stretch=0)

        spacer = Widgets.Label('')
        vbox.add_widget(spacer, stretch=1)

        top.add_widget(sw, stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(btns, stretch=0)

        container.add_widget(top, stretch=1)

    def close(self):
        if self.fitsimage is None:
            self.fv.stop_global_plugin(str(self))
        else:
            self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        pass

    def pause(self):
        self.stop_blinking()

    def resume(self):
        # Don't automatically resume blinking
        pass

    def stop(self):
        self.stop_blinking()

    def redo(self, *args):
        pass

    def _blink_timer_cb(self, timer):
        # set timer
        cur_time = time.time()
        deadline = cur_time + self.interval

        if self.blink_channels:
            self.fv.gui_do(self.fv.next_channel)
        else:
            self.fv.gui_do(self.fv.next_img, loop=True)

        timer.set(max(0.001, deadline - time.time()))

    def start_blinking(self):
        self.blink_timer.set(self.interval)

    def stop_blinking(self):
        self.blink_timer.clear()

    def _start_blink_cb(self):
        if not self.blink_channels:
            # Don't start blinking if there are no multiple images to blink
            # in this channel
            if len(self.channel) <= 1:
                self.fv.show_error(
                    "Blink opened in local mode and there are not multiple "
                    "images to blink in this channel")
                return

        self._set_interval_cb()

    def _stop_blink_cb(self):
        self.stop_blinking()

    def _set_interval_cb(self):
        interval = float(self.w.interval.get_text())
        self.interval = max(min(interval, self.ival_max), self.ival_min)
        self.stop_blinking()
        self.start_blinking()

    def _set_blink_mode_cb(self, tf):
        self.blink_channels = tf

    def __str__(self):
        return 'blink'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_Blink', package='ginga')

# END



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from astLib import astWCS, astCoords
# astlib requires pyfits (or astropy) in order
# to create a WCS object from a FITS header.
from astropy.io import fits as pyfits

from ginga.util.wcsmod import common

astWCS.NUMPY_MODE = True

coord_types = ['j2000', 'b1950', 'galactic']


class AstLibWCS(common.BaseWCS):
    """
    A WCS interface for astLib.astWCS.WCS
    You need to install python module 'astLib'
    (http://sourceforge.net/projects/astlib)
    if you want to use this version.

    """
    def __init__(self, logger):
        super(AstLibWCS, self).__init__(logger)

        self.kind = 'astlib/wcstools'

    def load_header(self, header, fobj=None):
        self.header = {}
        self.header.update(header.items())

        self.fix_bad_headers()

        try:
            # reconstruct a pyfits header
            hdr = pyfits.Header(header.items())

            self.logger.debug("Trying to make astLib wcs object")
            self.wcs = astWCS.WCS(hdr, mode='pyfits')

            self.coordsys = self.get_coord_system_name(self.header)
            self.logger.debug("Coordinate system is: %s" % (self.coordsys))

        except Exception as e:
            self.logger.error("Error making WCS object: %s" % (str(e)))
            self.wcs = None

    def get_coord_system_name(self, header):
        coordsys = common.get_coord_system_name(header)
        coordsys = coordsys.upper()
        if coordsys in ('FK4',):
            return 'b1950'
        elif coordsys in ('FK5', 'ICRS'):
            return 'j2000'
        elif coordsys in ('PIXEL',):
            return 'pixel'

        #raise common.WCSError("Cannot determine appropriate coordinate system from FITS header")  # noqa
        return 'j2000'

    def spectral_coord(self, idxs, coords='data'):
        raise common.WCSError("This feature not supported by astWCS")

    def pixtoradec(self, idxs, coords='data'):
        if coords == 'fits':
            # Via astWCS.NUMPY_MODE, we've forced pixels referenced from 0
            idxs = tuple(map(lambda x: x - 1, idxs))

        try:
            ra_deg, dec_deg = self.wcs.pix2wcs(idxs[0], idxs[1])

        except Exception as e:
            self.logger.error("Error calculating pixtoradec: %s" % (str(e)))
            raise common.WCSError(e)

        return ra_deg, dec_deg

    def radectopix(self, ra_deg, dec_deg, coords='data', naxispath=None):
        try:
            x, y = self.wcs.wcs2pix(ra_deg, dec_deg)

        except Exception as e:
            self.logger.error("Error calculating radectopix: %s" % (str(e)))
            raise common.WCSError(e)

        if coords == 'fits':
            # Via astWCS.NUMPY_MODE, we've forced pixels referenced from 0
            x, y = x + 1, y + 1

        return (x, y)

    def pixtosystem(self, idxs, system=None, coords='data'):

        if self.coordsys == 'raw':
            raise common.WCSError("No usable WCS")

        if system is None:
            system = 'j2000'

        # Get a coordinates object based on ra/dec wcs transform
        ra_deg, dec_deg = self.pixtoradec(idxs, coords=coords)
        self.logger.debug("ra, dec = %f, %f" % (ra_deg, dec_deg))

        # convert to alternate coord
        try:
            fromsys = self.coordsys.upper()

            if fromsys == 'PIXEL':
                # these are really pixel values
                return (ra_deg, dec_deg)

            tosys = system.upper()

            if fromsys == 'B1950':
                equinox = 1950.0
            else:
                equinox = 2000.0

            lon_deg, lat_deg = astCoords.convertCoords(fromsys, tosys,
                                                       ra_deg, dec_deg,
                                                       equinox)
        except Exception as e:
            raise common.WCSError(
                "Error converting between coordinate systems "
                "'%s' and '%s': %s" % (fromsys, tosys, str(e)))

        return (lon_deg, lat_deg)


# register our WCS with ginga
common.register_wcs('astlib', AstLibWCS, coord_types)



#
# contour.py -- Support for drawing contours on ginga canvases
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

# THIRD-PARTY
import numpy as np

have_skimage = False
try:
    from skimage import measure
    have_skimage = True
except ImportError:
    pass


def get_contours(data, levels):
    """Get sets of contour points for numpy array `data`.
    `levels` is a sequence of values around which to calculate the
    contours.  Returns a list of numpy arrays of points--each array
    makes a polygon if plotted as such.
    """
    if not have_skimage:
        raise Exception("Please install scikit-image > 0.13"
                        "to use this function")
    res = [filter(lambda arr: len(arr) >= 3,
                  measure.find_contours(data, level))
           for level in levels]
    res = [map(lambda arr: np.roll(arr, 1, axis=1),
               arrs) for arrs in res]
    return res


def calc_contours(data, num_contours):
    """Get sets of contour points for numpy array `data`.
    `num_contours` specifies the number (int) of contours to make.
    Returns a list of numpy arrays of points--each array makes a polygon
    if plotted as such.
    """
    mn = np.nanmean(data)
    top = np.nanmax(data)
    levels = np.linspace(mn, top, num_contours)
    return get_contours(data, levels)


def create_contours_obj(canvas, contour_groups, colors=None, **kwargs):
    """Create and return a compound object for ginga `canvas`, consisting
    of a number of contour polygons.  `contour_groups` is a list of
    numpy arrays of points representing polygons, such as returned by
    calc_contours() or get_contours().
    `colors` (if provided) is a list of colors for each polygon.
    Any other keyword parameters are passed on to the Polygon class.
    """
    if colors is None:
        colors = ['black']
    Polygon = canvas.get_draw_class('polygon')
    Compound = canvas.get_draw_class('compoundobject')
    objs = [Polygon(contour, color=colors[i % len(colors)], **kwargs)
            for i, contours in enumerate(contour_groups)
            for n, contour in enumerate(contours)
            ]
    contours_obj = Compound(*objs)
    return contours_obj

# END



# This contains imports plugins that configure pytest for Ginga tests.
# By importing them here in conftest.py they are discoverable by pytest
# no matter how it is invoked within the source tree.

from astropy.tests.helper import enable_deprecations_as_exceptions
from astropy.tests.plugins.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions
enable_deprecations_as_exceptions()

## Uncomment and customize the following lines to add/remove entries from
## the list of packages for which version numbers are displayed when running
## the tests. Making it pass for KeyError is essential in some cases when
## the package uses other Astropy affiliated packages.
try:
    PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
    del PYTEST_HEADER_MODULES['h5py']
except KeyError:
    pass

## Uncomment the following lines to display the version number of the
## package rather than the version number of Astropy in the top line when
## running the tests.
import os

## This is to figure out the affiliated package version, rather than
## using Astropy's
from . import version

try:
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version.version
except KeyError:
    pass



#
# AggHelp.py -- help classes for the Agg drawing
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import aggdraw as agg

from ginga import colors
from ginga.fonts import font_asst


def get_cached_font(fontname, fontsize, color, alpha):
    key = (fontname, fontsize, color, alpha)
    try:
        return font_asst.get_cache(key)

    except KeyError:
        # see if we can build the font
        info = font_asst.get_font_info(fontname, subst_ok=True)

        font = agg.Font(color, info.font_path, size=fontsize, opacity=alpha)
        font_asst.add_cache(key, font)

        return font


def load_font(font_name, font_file):
    # try to load it as a sanity check
    #agg.Font('black', font_file, size=10, opacity=255)

    if not font_asst.have_font(font_name):
        font_asst.add_font(font_file, font_name=font_name)
    return font_name


class AggContext(object):

    def __init__(self, canvas):
        self.canvas = canvas

    def set_canvas(self, canvas):
        self.canvas = canvas

    def get_color(self, color):
        if isinstance(color, str) or isinstance(color, type(u"")):
            r, g, b = colors.lookup_color(color)
        elif isinstance(color, tuple):
            # color is assumed to be a 3-tuple of RGB values as floats
            # between 0 and 1
            r, g, b = color
        else:
            r, g, b = 1.0, 1.0, 1.0

        return (int(r * 255), int(g * 255), int(b * 255))

    def get_pen(self, color, linewidth=1, alpha=1.0):
        # if hasattr(self, 'linestyle'):
        #     if self.linestyle == 'dash':
        #         cr.set_dash([ 3.0, 4.0, 6.0, 4.0], 5.0)
        op = int(alpha * 255)

        p = agg.Pen(self.get_color(color), width=linewidth,
                    opacity=op)
        return p

    def get_brush(self, color, alpha=1.0):
        op = int(alpha * 255)
        b = agg.Brush(self.get_color(color), opacity=op)
        return b

    def get_font(self, name, size, color, alpha=1.0):
        color = self.get_color(color)
        op = int(alpha * 255)

        name = font_asst.resolve_alias(name, name)
        font = get_cached_font(name, size, color, op)
        return font

    def text_extents(self, text, font):
        wd, ht = self.canvas.textsize(text, font)
        return wd, ht

# END



#! /usr/bin/env python
#
# example5_mpl.py -- Load a fits file into a Ginga widget with a
#          matplotlib backend.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
#
"""
   $ ./example5_mpl.py [fits file]

This example program shows to capture button and keypress events
for your own use.  After loading a FITS file use the following keys:

Press 'x' to turn on capture of events and bypass most normal keystroke
processing. Press it again to resume normal processing.  An on-screen
message will tell you which mode you are in.

While in 'capture mode' you can draw points with the right mouse button.
Press 'c' to clear the canvas of drawn points.
"""

import sys
import os
import platform
# just in case you want to use qt
os.environ['QT_API'] = 'pyqt'

import matplotlib
options = ['Qt4Agg', 'GTK', 'GTKAgg', 'MacOSX', 'GTKCairo', 'WXAgg',
           'TkAgg', 'QtAgg', 'FltkAgg', 'WX']
# Force a specific toolkit on mac
macos_ver = platform.mac_ver()[0]
if len(macos_ver) > 0:
    # change this to "pass" if you want to force a different backend
    # On Mac OS X I found the default choice for matplotlib is not stable
    # with ginga
    matplotlib.use('Qt4Agg')
import matplotlib.pyplot as plt

from ginga.mplw.ImageViewCanvasMpl import ImageViewCanvas
from ginga.mplw.ImageViewCanvasTypesMpl import DrawingCanvas
from ginga.misc import log
from ginga.util.loader import load_data

# Set to True to get diagnostic logging output
use_logger = False


class MyGingaFigure(object):
    def __init__(self, logger, fig):
        self.logger = logger
        # create a ginga object and tell it about the figure
        fi = ImageViewCanvas(logger=logger)
        fi.enable_autocuts('on')
        fi.set_autocut_params('zscale')
        fi.add_callback('key-press', self.key_press_ginga)
        fi.set_figure(fig)
        self.fitsimage = fi

        # enable all interactive features
        fi.get_bindings().enable_all(True)

        canvas = DrawingCanvas()
        canvas.enable_draw(True)
        canvas.set_callback('button-press', self.btn_down)
        canvas.set_callback('button-release', self.btn_up)
        canvas.set_drawtype('point', color='cyan')
        canvas.set_callback('draw-event', self.draw_event)
        canvas.add_callback('key-press', self.key_press)
        canvas.set_surface(self.fitsimage)
        canvas.ui_set_active(True)
        self.canvas = canvas

    def load(self, fitspath):
        # load an image
        image = load_data(fitspath, logger=self.logger)
        self.fitsimage.set_image(image)

    def capture(self):
        """
        Insert our canvas so that we intercept all events before they reach
        processing by the bindings layer of Ginga.
        """
        # insert the canvas
        self.fitsimage.add(self.canvas, tag='mycanvas')

    def release(self):
        """
        Remove our canvas so that we no longer intercept events.
        """
        # retract the canvas
        self.fitsimage.delete_object_by_tag('mycanvas')

    def clear(self):
        """
        Clear the canvas of any drawing made on it.
        """
        self.canvas.delete_all_objects()

    def get_wcs(self, data_x, data_y):
        """Return (re_deg, dec_deg) for the (data_x, data_y) position
        based on any WCS associated with the loaded image.
        """
        img = self.fitsimage.get_image()
        ra, dec = img.pixtoradec(data_x, data_y)
        return ra, dec

    # CALLBACKS
    # NOTE: return values on callbacks are important: if True then lower
    # layer Ginga canvas items will not get events

    def key_press(self, canvas, keyname):
        if keyname == 'x':
            self.fitsimage.onscreen_message("Moving to regular mode",
                                            delay=1.0)
            self.release()
        elif keyname == 'c':
            self.clear()
            return True

        fi = canvas.fitsimage
        data_x, data_y = fi.get_last_data_xy()
        ra, dec = self.get_wcs(data_x, data_y)
        print("key %s pressed at data %d,%d ra=%s dec=%s" % (
            keyname, data_x, data_y, ra, dec))
        return True

    def key_press_ginga(self, fitsimage, keyname):
        if keyname == 'x':
            self.fitsimage.onscreen_message("Moving to capture mode",
                                            delay=1.0)
            self.capture()
        return True

    def btn_down(self, canvas, button, data_x, data_y):
        ra, dec = self.get_wcs(data_x, data_y)
        print("button %s pressed at data %d,%d ra=%s dec=%s" % (
            button, data_x, data_y, ra, dec))
        return False

    def btn_up(self, canvas, button, data_x, data_y):
        ra, dec = self.get_wcs(data_x, data_y)
        print("button %s released at data %d,%d ra=%s dec=%s" % (
            button, data_x, data_y, ra, dec))
        return False

    def draw_event(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        data_x, data_y = obj.x, obj.y
        ra, dec = self.get_wcs(data_x, data_y)
        print("A %s was drawn at data %d,%d ra=%s dec=%s" % (
            obj.kind, data_x, data_y, ra, dec))
        return True


# create a regular matplotlib figure
fig = plt.figure()

# Here is our object
logger = log.get_logger(null=not use_logger, log_stderr=True)
foo = MyGingaFigure(logger, fig)

# load an image, if one was provided
if len(sys.argv) > 1:
    foo.load(sys.argv[1])

# Press 'x' to turn on capture of events. Press it again to resume normal
#   processing of events.
# Press 'c' to clear the canvas of drawn points.
plt.show()



#! /usr/bin/env python
#
# example_wireframe.py -- Example of a 3D plot with octahedron and wireframe
#
"""
Example of 3D plotting in Ginga

Plots an octahedron within a wireframe sphere.

Run with no parameters.  Scroll to zoom in/out, click and drag to orbit.

Requirements: Qt5, OpenGL, numpy
"""

import sys

import numpy as np

from ginga import toolkit
toolkit.use('qt5')

from ginga.gw import Widgets  # noqa
from ginga.opengl.ImageViewQtGL import CanvasView  # noqa
from ginga.canvas.CanvasObject import get_canvas_types  # noqa
from ginga.misc import log  # noqa


class Viewer(object):

    def __init__(self, app):
        super(Viewer, self).__init__()
        self.logger = app.logger
        self.dc = get_canvas_types()

        self.top = app.make_window(title="Simple Ginga 3D Viewer")

        vw = CanvasView(self.logger)
        vw.ui_set_active(True)
        self.vw = vw

        # quick hack to get 'u' to invoke hidden camera mode
        bm = vw.get_bindmap()
        bm.mode_map['u'] = bm.mode_map['mode_camera']

        bd = vw.get_bindings()
        bd.enable_all(True)

        # canvas that we will draw on
        canvas = self.dc.DrawingCanvas()
        canvas.set_surface(vw)
        self.canvas = canvas
        # add canvas to view
        private_canvas = vw.get_canvas()
        private_canvas.add(canvas)

        # add little mode indicator that shows keyboard modal states
        #vw.show_mode_indicator(True, corner='ur')

        # little hack because we don't have a way yet to ask for this
        # variation of back end through ginga.toolkit
        ww = Widgets.wrap(vw.get_widget())

        vbox = Widgets.VBox()
        vbox.set_border_width(2)
        vbox.set_spacing(1)
        vbox.add_widget(ww, stretch=1)

        hbox = Widgets.HBox()
        hbox.set_border_width(4)

        wquit = Widgets.Button("Quit")
        wquit.add_callback('activated', self.quit)
        hbox.add_widget(Widgets.Label(''), stretch=1)
        hbox.add_widget(wquit)

        vbox.add_widget(hbox)

        self.top.set_widget(vbox)

    def quit(self, w):
        self.top.delete()
        sys.exit(0)


def plot_octahedron(viewer, r):
    # octahedron
    A = [0.17770898, 0.72315927, 0.66742804]
    B = [-0.65327074, -0.4196453, 0.63018661]
    C = [0.65382635, 0.42081934, -0.62882604]
    D = [-0.17907021, -0.72084723, -0.66956189]
    E = [-0.73452809, 0.5495376, -0.39809158]
    F = [0.73451554, -0.55094017, 0.39617148]
    octo = [[E, A, B],
            [E, B, D],
            [E, D, C],
            [E, C, A],
            [F, A, B],
            [F, B, D],
            [F, D, C],
            [F, C, A],
            ]
    clrs = [('gray%d' % (i * 10 + 5)) for i in range(8)]
    for i, tri in enumerate(octo):
        new_tri = [np.asarray(pt) * r for pt in tri]
        viewer.canvas.add(viewer.dc.Polygon(new_tri, color='yellow',
                                            fill=True, fillcolor=clrs[i],
                                            fillalpha=0.4))


def get_wireframe(viewer, x, y, z, **kwargs):
    """Produce a compound object of paths implementing a wireframe.
    x, y, z are expected to be 2D arrays of points making up the mesh.
    """
    # TODO: something like this would make a great utility function
    # for ginga
    n, m = x.shape
    objs = []
    for i in range(n):
        pts = np.asarray([(x[i][j], y[i][j], z[i][j])
                          for j in range(m)])
        objs.append(viewer.dc.Path(pts, **kwargs))

    for j in range(m):
        pts = np.asarray([(x[i][j], y[i][j], z[i][j])
                          for i in range(n)])
        objs.append(viewer.dc.Path(pts, **kwargs))

    return viewer.dc.CompoundObject(*objs)


def plot_sphere(viewer, r):
    # sphere
    u = np.linspace(0, np.pi, 30)
    v = np.linspace(0, 2 * np.pi, 30)
    x = np.outer(np.sin(u), np.sin(v)) * r
    y = np.outer(np.sin(u), np.cos(v)) * r
    z = np.outer(np.cos(u), np.ones_like(v)) * r

    wf = get_wireframe(viewer, x, y, z, color='cyan', alpha=0.3)
    viewer.canvas.add(wf)


logger = log.get_logger('example', level=20, log_stderr=True)

app = Widgets.Application(logger)

v = Viewer(app)
v.top.resize(512, 512)
v.top.show()

# put viewer in camera mode
bm = v.vw.get_bindmap()
bm.set_mode('camera', mode_type='locked')

# toggle 3D view
bd = v.vw.get_bindings()
bd.kp_camera_toggle3d(v.vw, None, 0, 0)

r = 100
plot_octahedron(v, r)
plot_sphere(v, r)

app.mainloop()

# END



#
# CanvasRenderMpl.py -- for rendering into a ImageViewMpl widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import matplotlib.patches as patches
import matplotlib.lines as lines
from matplotlib.path import Path as MplPath

from . import MplHelp
from ginga.canvas.mixins import *  # noqa
from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)
        self.shape = None

        # TODO: encapsulate this drawable
        self.cr = MplHelp.MplContext(surface)

        self.pen = None
        self.brush = None
        self.font = None

    def set_line_from_shape(self, shape):
        # TODO: support line width and style
        alpha = getattr(shape, 'alpha', 1.0)
        linewidth = getattr(shape, 'linewidth', 1.0)
        linestyle = getattr(shape, 'linestyle', 'solid')
        self.pen = self.cr.get_pen(shape.color, linewidth=linewidth,
                                   linestyle=linestyle, alpha=alpha)

    def set_fill_from_shape(self, shape):
        fill = getattr(shape, 'fill', False)
        if fill:
            if hasattr(shape, 'fillcolor') and shape.fillcolor:
                color = shape.fillcolor
            else:
                color = shape.color
            alpha = getattr(shape, 'alpha', 1.0)
            alpha = getattr(shape, 'fillalpha', alpha)
            self.brush = self.cr.get_brush(color, alpha=alpha)
        else:
            self.brush = None

    def set_font_from_shape(self, shape):
        if hasattr(shape, 'font'):
            if hasattr(shape, 'fontsize') and shape.fontsize is not None:
                fontsize = shape.fontsize
            else:
                fontsize = shape.scale_font(self.viewer)
            fontsize = self.scale_fontsize(fontsize)
            alpha = getattr(shape, 'alpha', 1.0)
            self.font = self.cr.get_font(shape.font, fontsize, shape.color,
                                         alpha=alpha)
        else:
            self.font = None

    def initialize_from_shape(self, shape, line=True, fill=True, font=True):
        if line:
            self.set_line_from_shape(shape)
        if fill:
            self.set_fill_from_shape(shape)
        if font:
            self.set_font_from_shape(shape)

    def set_line(self, color, alpha=1.0, linewidth=1, style='solid'):
        # TODO: support style
        self.pen = self.cr.get_pen(color, alpha=alpha, linewidth=linewidth,
                                   linestyle=style)

    def set_fill(self, color, alpha=1.0):
        if color is None:
            self.brush = None
        else:
            self.brush = self.cr.get_brush(color, alpha=alpha)

    def set_font(self, fontname, fontsize, color='black', alpha=1.0):
        fontsize = self.scale_fontsize(fontsize)
        self.font = self.cr.get_font(fontname, fontsize, color,
                                     alpha=alpha)

    def text_extents(self, text):
        return self.cr.text_extents(text, self.font)

    ##### DRAWING OPERATIONS #####

    def draw_text(self, cx, cy, text, rot_deg=0.0):
        fontdict = self.font.get_fontdict()
        self.cr.push(allow=['alpha', 'color'])
        self.cr.set(rotation=rot_deg)

        self.cr.axes.text(cx, cy, text, fontdict=fontdict,
                          **self.cr.kwdargs)
        self.cr.pop()

    def draw_polygon(self, cpoints):
        self.cr.init(closed=True, transform=None)
        self.cr.update_patch(self.pen, self.brush)

        xy = trcalc.strip_z(cpoints)

        p = patches.Polygon(xy, **self.cr.kwdargs)
        self.cr.axes.add_patch(p)

    def draw_circle(self, cx, cy, cradius):
        self.cr.init(radius=cradius, transform=None)
        self.cr.update_patch(self.pen, self.brush)

        xy = (cx, cy)

        p = patches.Circle(xy, **self.cr.kwdargs)
        self.cr.axes.add_patch(p)

    def draw_bezier_curve(self, verts):
        self.cr.init(transform=None)
        self.cr.update_patch(self.pen, None)

        codes = [MplPath.MOVETO,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 ]
        path = MplPath(verts, codes)

        p = patches.PathPatch(path, **self.cr.kwdargs)
        self.cr.axes.add_patch(p)

    def draw_ellipse_bezier(self, verts):
        self.cr.init(transform=None)
        self.cr.update_patch(self.pen, self.brush)

        # draw 4 bezier curves to make the ellipse
        codes = [MplPath.MOVETO,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 ]
        path = MplPath(verts, codes)

        p = patches.PathPatch(path, **self.cr.kwdargs)
        self.cr.axes.add_patch(p)

    def draw_line(self, cx1, cy1, cx2, cy2):
        self.cr.init(transform=None)
        self.cr.update_line(self.pen)

        l = lines.Line2D((cx1, cx2), (cy1, cy2), **self.cr.kwdargs)
        self.cr.axes.add_line(l)

    def draw_path(self, cpoints):
        self.cr.init(closed=False, transform=None)
        self.cr.update_patch(self.pen, None)

        xy = trcalc.strip_z(cpoints)

        p = patches.Polygon(xy, **self.cr.kwdargs)
        self.cr.axes.add_patch(p)


class CanvasRenderer(render.RendererBase):

    def __init__(self, viewer):
        render.RendererBase.__init__(self, viewer)

        self.kind = 'mpl'
        self.rgb_order = viewer.rgb_order
        self.surface = None

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        pass

    def render_image(self, rgbobj, dst_x, dst_y):
        # for compatibility with the other renderers
        return self.viewer.render_image(rgbobj, dst_x, dst_y)

    def get_surface_as_array(self, order=None):
        raise render.RenderError("This renderer can only be used with a matplotlib viewer")

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.viewer.ax_util)
        cr.initialize_from_shape(shape)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        cr.set_font_from_shape(shape)
        return cr.text_extents(shape.text)


#END
